__all__ = ('shield',)

from .future import Future


class _FutureChainer:
    """
    Chains a future's result into an other one used as a callback of the source future.
    
    Attributes
    ----------
    target : ``Future``
        The target future to chain the result into if applicable.
    """
    __slots__ = ('target',)
    
    def __init__(self, target):
        """
        Creates a new ``_FutureChainer`` with the given target future.
        
        Parameters
        ----------
        target : ``Future``
            The target future to chain the result into if applicable.
        """
        self.target = target
    
    def __call__(self, future):
        """
        Chains the source future's result into the target one.
        
        Parameters
        ----------
        future : ``Future``
            The source future to chain it's result from.
        """
        # remove chain remover
        target = self.target
        callbacks = target._callbacks
        for index in range(len(callbacks)):
            callback = callbacks[index]
            if (type(callback) is _ChainRemover) and (callback.target is future):
                del callbacks[index]
                break
        
        # set result
        if target.is_pending():
            target._state |= future._state
            target._result = future._result
            
            target._loop._schedule_callbacks(target)


class _ChainRemover:
    """
    Removes the ``_FutureChainer`` callback of the source future if the chained target future is marked as done before.
    
    Parameters
    -------
    target : ``Future``
        The source future.
    """
    __slots__ = ('target',)
    
    def __init__(self, target):
        self.target = target
    
    def __call__(self, future):
        # remove chainer
        callbacks = self.target._callbacks
        for index in range(len(callbacks)):
            callback = callbacks[index]
            if (type(callback) is _FutureChainer) and (callback.target is future):
                del callbacks[index]
                # because this might be the only place, where we retrieve the result, we will just silence it.
                future.silence()
                break


def shield(awaitable, loop):
    """
    Protects the given `awaitable` from being cancelled.
    
    Parameters
    ----------
    awaitable : `awaitable`
        The awaitable object to shield.
    loop : ``EventThread``
        The event loop to run the `awaitable` on.
    
    Returns
    -------
    un_protected : ``Future``
        If the given `awaitable` is a finished task, returns it, else returns a ``Future``, to what the original
        awaitable's result will be chained to.
    
    Raises
    -------
    TypeError
        The given `awaitable` is not `awaitable`.
    """
    protected = loop.ensure_future(awaitable)
    if protected.is_done():
        return protected # already done, we can return
    
    un_protected = Future(loop)
    protected._callbacks.append(_FutureChainer(un_protected))
    un_protected._callbacks.append(_ChainRemover(protected))
    return un_protected



import sys, warnings
from types import ModuleType

from . import asyncio

try:
    module = sys.modules['asyncio']
except KeyError:
    module = asyncio
    sys.modules['asyncio'] = asyncio

else:
    warnings.warn(f'asyncio is already loaded, replacing it.')
    
    try:
        del sys.modules['_asyncio']
    except KeyError:
        pass
    
    module.__dict__.clear()
    module.__dict__.update(asyncio.__dict__)
    module.__file__ = __file__


for sub_module_name, feature_names, extra_features in (
    ('base_events', (
        'BaseEventLoop',
        '_run_until_complete_cb', # Required by anyio
        ), None,
    ),
    ('base_futures', None, None),
    ('base_subprocess', None, None),
    ('base_tasks', None, None),
    ('constants', None, None),
    ('coroutines', (
        'coroutine',
        'iscoroutinefunction',
        'iscoroutine',
        ), (
        ('_DEBUG', False), # Required by aiohttp 3.8
        )
    ),
    ('events', (
        'AbstractEventLoopPolicy',
        'AbstractEventLoop',
        'AbstractServer',
        'Handle',
        'TimerHandle',
        'get_event_loop_policy',
        'set_event_loop_policy',
        'get_event_loop',
        'set_event_loop',
        'new_event_loop',
        'get_child_watcher',
        'set_child_watcher',
        '_set_running_loop',
        'get_running_loop',
        '_get_running_loop',
        ), None,
    ),
    ('exceptions', (
        'CancelledError',
        'InvalidStateError',
        'TimeoutError',
        'IncompleteReadError',
        'LimitOverrunError',
        'SendfileNotAvailableError',
        ), None,
    ),
    ('format_helpers', None, None),
    ('futures', (
        'Future',
        'wrap_future',
        'isfuture',
        ), None,
    ),
    ('locks', (
        'Lock',
        'Event',
        'Condition',
        'Semaphore',
        'BoundedSemaphore',
        ), None,
    ),
    ('proactor_events', (
        'BaseProactorEventLoop',
        ), None,
    ),
    ('protocols', (
        'BaseProtocol',
        'Protocol',
        'DatagramProtocol',
        'SubprocessProtocol',
        'BufferedProtocol',
        ), None,
    ),
    ('queues', (
        'Queue',
        'PriorityQueue',
        'LifoQueue',
        'QueueFull',
        'QueueEmpty',
        ), None,
    ),
    ('runners', (
        'run',
        ), None,
    ),
    ('selector_events', (
        'BaseSelectorEventLoop',
        ), None,
    ),
    ('sslproto', None, None),
    ('staggered_race', (
        'staggered_race',
        ), None,
    ),
    ('streams', (
        'StreamReader',
        'StreamWriter',
        'StreamReaderProtocol',
        'open_connection',
        'start_server',
        'start_unix_server',
        ), None,
    ),
    ('subprocess', (
        'DEVNULL',
        'PIPE',
        'Process', # Required by anyio
        'STDOUT',
        'create_subprocess_exec',
        'create_subprocess_shell',
        ), None,
    ),
    ('tasks', (
        'Task',
        'create_task',
        'FIRST_COMPLETED',
        'FIRST_EXCEPTION',
        'ALL_COMPLETED',
        'wait',
        'wait_for',
        'as_completed',
        'sleep',
        'gather',
        'shield',
        'ensure_future',
        'run_coroutine_threadsafe',
        'current_task',
        'all_tasks',
        '_register_task',
        '_unregister_task',
        '_enter_task',
        '_leave_task',
        ), None,
    ),
    ('threads', (
        'to_thread',
        ), None,
    ),
    ('transports', (
        'BaseTransport',
        'ReadTransport',
        'WriteTransport',
        'Transport',
        'DatagramTransport',
        'SubprocessTransport',
        ), None,
    ),
    ('trsock', None, None),
    ('unix_events', (
        'SelectorEventLoop',
        'AbstractChildWatcher',
        'SafeChildWatcher',
        'FastChildWatcher',
        'PidfdChildWatcher',
        'MultiLoopChildWatcher',
        'ThreadedChildWatcher',
        'DefaultEventLoopPolicy',
        ), None,
    ),
    ('windows_events', (
        'SelectorEventLoop',
        'ProactorEventLoop',
        'IocpProactor',
        'DefaultEventLoopPolicy',
        'WindowsSelectorEventLoopPolicy',
        'WindowsProactorEventLoopPolicy',
        ), None,
    ),
    ('windows_utils', (
        'pipe',
        'Popen',
        'PIPE',
        'PipeHandle',
        ), None,
    )
):
    
    module_name = f'asyncio.{sub_module_name}'
    
    try:
        module = sys.modules[module_name]
    except KeyError:
        module = ModuleType(module_name)
        sys.modules[module_name] = module
    else:
        module.__dict__.clear()
    
    asyncio.__dict__[sub_module_name] = module
    module.__file__ = __file__
    
    if (feature_names is not None):
        module.__all__ = feature_names
        module.__dict__.update((feature_name, getattr(asyncio, feature_name)) for feature_name in feature_names)
    
    if (extra_features is not None):
        module.__dict__.update(extra_features)

del extra_features
del sub_module_name
del feature_names
del module_name
del module


from . asyncio import *
__all__ = asyncio.__all__


from .. import register_library_extension
register_library_extension('HuyaneMatsu.asyncio')
del register_library_extension



__all__ = ()

from ..rich_attribute_error import RichAttributeErrorBaseType


class ProxyBase(RichAttributeErrorBaseType):
    """
    Base type for object proxies.
    
    Attributes
    ----------
    _overwrites : `None`, `dict` of (`str`, `object`) items
        Field overwrites.
    _proxied : `object`
        The proxied object.
    """
    __slots__ = ('_overwrites', '_proxied')
    
    def __new__(cls, to_proxy):
        """
        Creates a new proxy object.
        
        Parameters
        ----------
        to_proxy : `instance<cls>`, `object`
            The object to proxy.
        """
        if isinstance(to_proxy, cls):
            return to_proxy._copy_self_deep()
        
        self = object.__new__(cls)
        self._overwrites = None
        self._proxied = to_proxy
        return self
    
    
    def _copy_self_deep(self):
        """
        Copies self with also copying the overwrites.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        overwrites = self._overwrites
        if (overwrites is not None):
            overwrites = overwrites.copy()
        new._overwrites = overwrites
        new._proxied = self._proxied
        return new
    
    
    def _copy_self_clean(self):
        """
        Copies self without copying the overwrites.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new._overwrites = None
        new._proxied = self._proxied
        return new
    
    
    def _set_overwrite(self, name, value):
        """
        Sets an overwrite for the given name.
        
        Parameters
        ----------
        name : `str`
            Attribute name.
        value : `object`
            Attribute value.
        """
        overwrites = self._overwrites
        if overwrites is None:
            overwrites = {}
            self._overwrites = overwrites
        
        overwrites[name] = value
    
    
    def _get_overwrite(self, name):
        """
        Gets an overwrite for the given name.
        
        Returns
        -------
        name : `str`
            The attribute's name.
        
        Returns
        -------
        found : `bool`
            Whether the field has an overwrite.
        value : `object`
            Field value. Defaults to `None` if not found.
        """
        overwrites = self._overwrites
        if overwrites is None:
            found = False
            value = None
        
        else:
            try:
                value = overwrites[name]
            except KeyError:
                found = False
                value = None
            else:
                found = True
        
        return found, value
    
    
    def _get_value(self, name):
        """
        Gets value for the given name.
        
        Returns
        -------
        name : `str`
            The attribute's name.
        
        Returns
        -------
        value : `object`
            Attribute value.
        """
        found, value = self._get_overwrite(name)
        if not found:
            value = getattr(self._proxied, name)
        
        return value
    
    
    def _del_overwrite(self, name):
        """
        Deletes an overwrite for the given name.
        
        Parameters
        ----------
        name : `str`
            Attribute name.
        """
        overwrites = self._overwrites
        if overwrites is not None:
            try:
                del overwrites[name]
            except KeyError:
                pass
            else:
                if not overwrites:
                    self._overwrites = None
    
    
    def __repr__(self):
        """Returns the proxy's representation."""
        return f'<{type(self).__name__} to {self._proxied!r}>'
    
    
    def __hash__(self):
        return hash(self._proxied)
    
    
    def __eq__(self, other):
        """Returns whether the two proxies are equal."""
        self_value = self._proxied
        
        if isinstance(other, type(self)):
            other_value = other._proxied
        else:
            other_value = other
        
        return self_value.__eq__(other_value)



__all__ = ('enter_executor',)

from threading import current_thread

from ...utils import ignore_frame, include

from ..exceptions import CancelledError

from .future import Future
from .future_states import (
    FUTURE_STATE_CANCELLED, FUTURE_STATE_CANCELLING_SELF, FUTURE_STATE_RESULT_RAISE, FUTURE_STATE_RESULT_RETURN
)
from .task import Task


FUTURE_MASK_CANCEL_INITIALISED_BEFORE = FUTURE_STATE_CANCELLING_SELF | FUTURE_STATE_CANCELLED
FUTURE_MASK_CANCEL_WITH_EXCEPTION = FUTURE_STATE_RESULT_RAISE | FUTURE_STATE_CANCELLED


ignore_frame(__spec__.origin, 'get_result', 'raise exception',)
ignore_frame(__spec__.origin, '__iter__', 'yield self',)

EventThread = include('EventThread')


class enter_executor:
    """
    Async context manager for moving a ``Task``'s section's execution to an executor thread.
    
    Usage
    -----
    ```py
    # Do async code
    async with enter_executor():
        # Do blocking stuff
        # Also async operation are still possible, but not recommended, because they cause my thread switch.
    
    # Do async code
    ```
    
    Attributes
    ----------
    _enter_future : `None`, ``Future``
        The future, what blocks the task's the execution, meanwhile the thread switch is taking place.
    _exit_future : `None`, ``Future``
        The future, what blocks the task's execution, meanwhile the thread is switching back.
    _waited_future : `None`, ``Future``
        Blocking future used inside of the task, meanwhile it is in executor.
    _task : `None`, ``Task``
    """
    __slots__ = ('_enter_future', '_exit_future', '_waited_future', '_task')
    
    def __init__(self):
        self._enter_future = None
        self._task = None
        self._exit_future = None
        self._waited_future = None
    
    
    async def __aenter__(self):
        """
        Moves the current tasks' execution to an executor thread.
        
        This method is a coroutine.
        
        Raises
        ------
        RuntimeError
            - Called from outside of an ``EventThread``.
            - Called from outside of a ``Task``.
        """
        thread = current_thread()
        if not isinstance(thread, EventThread):
            raise RuntimeError(
                f'`{self.__class__.__name__}` entered outside of `{EventThread.__name__}`, at {thread!r}.'
            )
        
        task = thread.current_task
        if task is None:
            raise RuntimeError(
                f'`{self.__class__.__name__}` entered outside of a `{Task.__name__}`.'
            )
        
        self._task = task
        loop = task._loop
        future = Future(loop)
        self._enter_future = future
        loop.call_soon(self._enter_executor)
        await future
        return self
    
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Moves the current task's executor back from an executor thread.
        
        This method is a coroutine.
        """
        await self._exit_future
        self._enter_future = None
        self._task = None
        self._exit_future = None
        self._waited_future = None
        return False
    
    
    def _enter_executor(self):
        """
        Moves the task's execution to an executor thread and wakes it up.
        """
        callbacks = self._enter_future._callbacks
        callbacks.clear()
        
        task = self._task
        task.add_done_callback(self._cancel_callback)
        
        task._loop.run_in_executor(self._executor_task)
    
    
    def _cancel_callback(self, future):
        """
        Callback added to the wrapped task. If the wrapped task is cancelled, then the section running inside of the
        executor will be cancelled as well, whenever it gives back the context with an `await`.
        """
        if future.is_cancelled():
            return
        
        waited_future = self._waited_future
        if waited_future is None:
            return
        
        waited_future.cancel()
    
    
    def _executor_task(self):
        """
        Wraps the tasks' section's running inside of an executor, still allowing it to use `await`-s.
        """
        task = self._task
        # relink future task
        loop = task._loop
        end_future = Future(loop)
        task._waited_future = end_future
        self._exit_future = end_future
        
        # Set result to the enter task, so it can be retrieved.
        self._enter_future.set_result(None)
        
        coroutine = task._coroutine
        
        # If some1 await at the block, we will sync_wrap it. If the exit future is awaited, then we quit.
        local_waited_future = None
        
        try:
            while True:
                if (local_waited_future is not None):
                    if local_waited_future is end_future:
                        end_future.set_result(None)
                        loop.call_soon_thread_safe(task._step)
                        break
                    
                    self._waited_future = local_waited_future
                    if task._state & FUTURE_MASK_CANCEL_INITIALISED_BEFORE:
                        local_waited_future.cancel()
                    
                    try:
                        local_waited_future.sync_wrap().wait_for_completion()
                    finally:
                        local_waited_future = None
                
                if task.is_done():
                    # there is no reason to raise
                    break
                
                try:
                    # Call either coroutine.throw(err) or coroutine.send(None).
                    if task._state & FUTURE_STATE_CANCELLING_SELF:
                        result = coroutine.throw(CancelledError())
                    else:
                        result = coroutine.send(None)
                
                except StopIteration as retrieved_exception:
                    state = task._state
                    if state & FUTURE_STATE_CANCELLING_SELF:
                        state |= FUTURE_STATE_CANCELLED
                        task._loop._schedule_callbacks(self)
                    else:
                        task._result = retrieved_exception.value
                        state |= FUTURE_STATE_RESULT_RETURN
                    
                    task._state = state
                    loop._schedule_callbacks(self)
                    loop.wake_up()
                    break
                    
                except CancelledError as retrieved_exception:
                    task._state |= FUTURE_MASK_CANCEL_WITH_EXCEPTION
                    
                    result = task._result
                    if (result is None):
                        task._result = retrieved_exception
                    else:
                        result.__cause__ = retrieved_exception
                    
                    loop._schedule_callbacks(self)
                    loop.wake_up()
                    break
                
                except BaseException as retrieved_exception:
                    task._result = retrieved_exception
                    task._state |= FUTURE_STATE_RESULT_RAISE
                    loop._schedule_callbacks(self)
                    loop.wake_up()
                    break
                
                else:
                    if isinstance(result, Future) and result._blocking:
                        result._blocking = False
                        local_waited_future = result
                        task._waited_future = result
                        
                        state = task._state
                        if state & FUTURE_STATE_CANCELLING_SELF:
                            if result.cancel():
                                task._state = state & ~FUTURE_STATE_CANCELLING_SELF
                    else:
                        continue
        
        finally:
            task.remove_done_callback(self._cancel_callback)
            self = None
            task = None



__all__ = ()


def proxy_property_factory(attribute_name):
    """
    Creates a new proxy property for the given attribute name.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute's name.
    
    Returns
    -------
    proxy_property : `property`
    """
    def getter(self):
        nonlocal attribute_name
        return self._get_value(attribute_name)
    
    
    def setter(self, value):
        nonlocal attribute_name
        self._set_overwrite(attribute_name, value)
    
    
    def deller(self):
        nonlocal attribute_name
        self._del_overwrite(attribute_name)
    
    return property(getter, setter, deller)


def proxy_method_factory(attribute_name):
    """
    Crates a new proxy method for the given attribute name.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute's name.
    
    Returns
    -------
    proxy_method : `FunctionType`
    """
    def proxy_method(self, *position_parameters, **keyword_parameters):
        method = self._get_value(attribute_name)
        return method(*position_parameters, **keyword_parameters)
    
    proxy_method.__name__ = attribute_name
    return proxy_method



__all__ = ('CoroutineFunctionTypeProxy',)

from types import FunctionType, MethodType

from ..async_utils import is_coroutine_function

from .coroutine_type import CoroutineTypeProxy
from .field_factories import proxy_method_factory, proxy_property_factory
from .proxy_base import ProxyBase


class CoroutineFunctionTypeProxy(ProxyBase):
    """
    Coroutine function proxy.
    
    Attributes
    ----------
    _overwrites : `None`, `dict` of (`str`, `object`) items
        Field overwrites.
    _proxied : `CoroutineFunctionType`
        The proxied coroutine function.
    """
    __slots__ = ()
    
    def __new__(cls, coroutine_function):
        """
        Creates a new coroutine function proxy.
        
        Parameters
        ----------
        to_proxy : `instance<cls>`, `CoroutineFunctionType`
            The coroutine function ot proxy.
        """
        if isinstance(coroutine_function, cls):
            return coroutine_function._copy_self_deep()
        
        if is_coroutine_function(coroutine_function):
            self = object.__new__(cls)
            self._proxied = coroutine_function
            self._overwrites = None
            return self
        
        raise TypeError(
            f'`coroutine_function` can be either a coroutine function or `{cls.__name__}`, '
            f'got {coroutine_function.__class__.__name__}.'
        )
    
    
    def __get__(self, instance, instance_type):
        if instance is None:
            return self
        
        return MethodType(self.__call__, instance)
    
    
    def __call__(self, *positional_parameters, **keyword_parameters):
        return CoroutineTypeProxy(self._proxied(*positional_parameters, **keyword_parameters))
    
    
    __qualname__ = FunctionType.__qualname__
    
    __annotations__ = proxy_property_factory('__annotations__')
    __class__ = proxy_property_factory('__class__')
    __closure__ = proxy_property_factory('__closure__')
    __code__ = proxy_property_factory('__code__')
    __defaults__ = proxy_property_factory('__defaults__')
    __dict__ = proxy_property_factory('__dict__')
    __doc__ = proxy_property_factory('__doc__')
    __globals__ = proxy_property_factory('__globals__')
    __kwdefaults__ = proxy_property_factory('__kwdefaults__')
    __name__ = proxy_property_factory('__name__')
    
    
    __reduce__ = proxy_method_factory('__reduce__')
    __reduce_ex__ = proxy_method_factory('__reduce_ex__')
    __sizeof__ = proxy_method_factory('__sizeof__')



__all__ = ('TaskGroup',)

from ...utils import RichAttributeErrorBaseType

from .future import Future
from .future_states import FUTURE_STATE_CANCELLED, FUTURE_STATE_RESULT_RAISE
from .task import Task


FUTURE_STATE_CANCELLED_OR_RAISE = FUTURE_STATE_CANCELLED | FUTURE_STATE_RESULT_RAISE


def _context_manager_leaver_cancel_on_exception(task_group, exception):
    """
    Context manager leaver that cancels the task of the given task group on exception.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The wrapped task group by the context manager.
    exception : `None`, `BaseException`
        The occurred exception if any.
    
    Returns
    -------
    captured : `bool`
        Whether the exception was captured.
    """
    if (exception is not None):
        task_group.cancel_all()
    
    return False


class TaskGroupContextManager(RichAttributeErrorBaseType):
    """
    Task group context manager.
    
    Attributes
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    leaver : `FunctionType`
        Function to call when the context is left.
        
        Can be implemented as:
        - `(Taskgroup, None | BaseException) -> bool`
    """
    __slots__ = ('leaver', 'task_group')
    
    def __new__(cls, task_group, leaver):
        """
        Creates a new task group context manager.
        
        Parameters
        ----------
        task_group : ``TaskGroup``
            The parent task group.
        leaver : `FunctionType`
            Function to call when the context is left.
        """
        self = object.__new__(cls)
        self.leaver = leaver
        self.task_group = task_group
        return self
    
    
    def __enter__(self):
        """
        Enters the context manager.
        
        Returns
        -------
        task_group : ``TaskGroup``
        """
        return self.task_group
    
    
    def __exit__(self, exception_type, exception, exception_traceback):
        """
        Exits the context manager.
        
        Parameters
        ----------
        exception_type : `None`, `type<BaseException>`
            The occurred exception's type if any.
        exception : `None`, `BaseException`
            The occurred exception if any.
        exception_traceback : `None`, `TracebackType`
            the exception's traceback if any.
        
        Returns
        -------
        captured : `bool`
            Whether the exception was captured.
        """
        return self.leaver(self.task_group, exception)
    
    
    def __repr__(self):
        """Returns the task group context manager representation."""
        return f'<{self.__class__.__name__} task_group = {self.taskk_group}, leaver = {self.leaver.__name__}>'


def _handler_wait_next():
    """
    Task group handler waiting for the next finishing task.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    Returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    result : `Nobe`
        Yes, we return `None`.
    """
    yield 1
    return 1, None


def _handler_wait_first(task_group):
    """
    Task group handler waiting for the first task to finish.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    Returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    result : ``Future``
        The first done task.
    """
    done = task_group.done
    if done:
        return 1, next(iter(done))
    
    if task_group.pending:
        return 1, (yield 1)
    
    return 1, None


def _handler_wait_first_and_pop(task_group):
    """
    Task group handler waiting for the first task to finish.
    
    The finished task will be popped from the task group's `.done`.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    Returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    result : ``Future``
        The first done task.
    """
    done = task_group.done
    if done:
        return 1, done.pop()
    
    if task_group.pending:
        return 0, (yield 0)
    
    return 1, None
    

def _handler_wait_exception(task_group):
    """
    Task group handler waiting for the first task to finish with an exception.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    result : ``Future``
        The first failing task.
    """
    for future in task_group.done:
        if future._state & FUTURE_STATE_RESULT_RAISE:
            return 1, future
    
    pending = task_group.pending
    while True:
        if not pending:
            return 1, None
        
        future = (yield 1)
        if future._state & FUTURE_STATE_RESULT_RAISE:
            return 1, future


def _handler_wait_exception_and_pop(task_group):
    """
    Task group handler waiting for the first task to finish with an exception.
    
    The finished task will be popped from the task group's `.done`.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    result : ``Future``
        The first failing task.
    """
    for future in task_group.done:
        if future._state & FUTURE_STATE_RESULT_RAISE:
            task_group.done.discard(future)
            return 0, future
    
    pending = task_group.pending
    while True:
        if not pending:
            return 1, None
        
        future = (yield 1)
        if future._state & FUTURE_STATE_RESULT_RAISE:
            return 0, future


def _handler_wait_exception_or_cancellation(task_group):
    """
    Task group handler waiting for the first task to finish with an exception or with cancellation.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    result : ``Future``
        The first failing task.
    """
    for future in task_group.done:
        if future._state & FUTURE_STATE_CANCELLED_OR_RAISE:
            return 1, future
    
    pending = task_group.pending
    while True:
        if not pending:
            return 1, None
        
        future = (yield 1)
        if future._state & FUTURE_STATE_CANCELLED_OR_RAISE:
            return 1, future


def _handler_wait_exception_or_cancellation_and_pop(task_group):
    """
    Task group handler waiting for the first task to finish with an exception or cancellation.
    
    The finished task will be popped from the task group's `.done`.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    result : ``Future``
        The first failing task.
    """
    for future in task_group.done:
        if future._state & FUTURE_STATE_CANCELLED_OR_RAISE:
            task_group.done.discard(future)
            return 0, future
    
    pending = task_group.pending
    while True:
        if not pending:
            return 1, None
        
        future = (yield 1)
        if future._state & FUTURE_STATE_CANCELLED_OR_RAISE:
            return 0, future


def _handler_wait_first_n(task_group, required_count):
    """
    Task group handler waiting for the first `0` task to finish.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    Returns
    -------
    done_count : `int`
        The amount of done futures.
    """
    done_count = len(task_group.done)
    while True:
        if done_count >= required_count:
            return 1, done_count
        
        (yield 1)
        done_count += 1


def _handler_wait_all(task_group):
    """
    Task group handler waiting for all tasks to finish.
    
    Parameters
    ----------
    task_group : ``TaskGroup``
        The parent task group.
    
    Yield
    -----
    should_add_to_done : `int`
        Whether the task should be added to done.
    
    Returns
    -------
    should_add_to_done : `int`
        Whether the task should be added to done.
    dummy : `None`
        Dummy value.
    """
    pending = task_group.pending
    while True:
        if not pending:
            return 1, None
        
        yield 1


class TaskGroup(RichAttributeErrorBaseType):
    """
    Represents grouped tasks (actually futures) on which it is possible to execute shared operations.
    
    Attributes
    ----------
    done : `set` of ``Future``
        The done tasks.
    loop : ``EventThread``
        The event loop used.
    pending : `set` pf ``Future``
        The pending tasks.
    waiters : `dict` of (``Future``, ``Generator``)
        Contains the active waiters with their handlers.
    """
    __slots__ = ('done', 'loop', 'pending', 'waiters')
    
    def __new__(cls, loop, tasks = None):
        """
        Creates a new task group.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to use.
        tasks : `None`, `iterable` of ``Future``
            Tasks (actually futures) to create the group with. 
        """
        done = set()
        pending = set()
        
        if (tasks is not None):
            for task in tasks:
                (done if task.is_done() else pending).add(task)
        
        self = object.__new__(cls)
        self.done = done
        self.loop = loop
        self.pending = pending
        self.waiters = {}
        
        if pending:
            for task in pending:
                task.add_done_callback(self._waited_done_callback)
        
        return self
    
    
    def __repr__(self):
        """Returns the task group's representation."""
        return f'<{self.__class__.__name__} done = {self.done!r}, pending = {self.pending!r}>'
    
    
    def _waited_done_callback(self, future):
        """
        Called when a waited future is done.
        
        Moves the future into the done ones and
        
        Parameters
        ----------
        future : ``Future``
            An added future.
        """
        self.pending.discard(future)
    
        done_waiters = None
        should_add_to_done = 1
        
        for waiter, handler in self.waiters.items():
            try:
                should_add_to_done &= handler.send(future)
            except StopIteration as err:
                result = err.value
                should_add_to_done &= result[0]
                waiter.set_result_if_pending(result[1])
                
                if done_waiters is None:
                    done_waiters = []
                done_waiters.append(waiter)
            
            except BaseException as err:
                waiter.set_exception_if_pending(err)
                
                if done_waiters is None:
                    done_waiters = []
                done_waiters.append(waiter)
        
        if should_add_to_done:
            self.done.add(future)
        
        if (done_waiters is not None):
            for waiter in done_waiters:
                try:
                    del self.waiters[waiter]
                except KeyError:
                    pass
    
    
    def _waiter_done_callback(self, future):
        """
        Called when a waiter future is done.
        
        Closes the waiter's handler.
        
        Parameters
        ----------
        future : `bool`
        """
        try:
            handler = self.waiters.pop(future)
        except KeyError:
            pass
        else:
            handler.close()
    
    
    def _add_handler(self, handler):
        """
        Adds a new handler to the task group.
        
        Parameters
        ----------
        handler : `GeneratorType`
            Handler to use.
        
        Returns
        -------
        waiter : ``Future``
        """
        waiter = Future(self.loop)
        
        try:
            handler.send(None)
        except StopIteration as err:
            waiter.set_result_if_pending(err.value[1])
        except BaseException as err:
            waiter.set_exception_if_pending(err)
        else:
            self.waiters[waiter] = handler
            waiter.add_done_callback(self._waiter_done_callback)
        
        return waiter
    
    # Registering
    
    def add_future(self, future):
        """
        Adds a new future (or task) to the task group.
        
        Parameters
        ----------
        future : ``Future``
            The future to add.
        
        Returns
        -------
        future : ``Future``
        
        Examples
        --------
        ```py3
        task_group = TaskGroup(loop)
        task_group.add_future(Task(loop, coro(12)))
        task_group.add_future(Future(loop))
        
        # Equals to:
        
        task_group = TaskGroup(loop, [Task(loop, coro(12)), Future(loop)])
        ```
        """
        if future.is_done():
            self._waited_done_callback(future)
        else:
            future.add_done_callback(self._waited_done_callback)
            self.pending.add(future)
        
        return future
    
    
    add_task = add_future
    
    
    def add_awaitable(self, coroutine_or_future):
        """
        Adds the given coroutine or future to the task group.
        
        Parameters
        ----------
        coroutine_or_future : `awaitable`
            The coroutine or future to add.
        
        Returns
        -------
        future : ``Future``
        
        Raises
        ------
        TypeError
            If `coroutine_or_future` is not `awaitable`.
        
        Examples
        --------
        ```py3
        task_group = TaskGroup(loop)
        task_group.add_awaitable(coro(12))
        task_group.add_awaitable(Future(loop))
        
        # Equals to:
        
        task_group = TaskGroup(loop, [Task(loop, coro(12)), Future(loop)])
        ```
        """
        return self.add_future(self.loop.ensure_future(coroutine_or_future))
    
    
    def create_future(self):
        """
        Creates a new future and adds it to the task group.
        
        Returns
        -------
        future : ``Future``
        
        Examples
        --------
        ```py3
        task_group = TaskGroup(loop)
        future = task_group.create_future()
        
        # Equals to:
        
        task_group = TaskGroup(loop, [Future(loop)])
        ```
        """
        future = Future(self.loop)
        future.add_done_callback(self._waited_done_callback)
        self.pending.add(future)
        return future
    
    
    def create_task(self, coroutine):
        """
        Creates a new task and adds it to the task group.
        
        Parameters
        ----------
        coroutine : ``Coroutine``, ``Generator``
            The coroutine to create task with.
        
        Returns
        -------
        task : ``Task``
        
        Examples
        --------
        ```py3
        task_group = TaskGroup(loop)
        task = task_group.create_task(coro(11))
        
        # Equals to:
        
        task_group = TaskGroup(loop, [Task(loop, coro(11))])
        ```
        """
        task = Task(self.loop, coroutine)
        task.add_done_callback(self._waited_done_callback)
        self.pending.add(task)
        return task
    
    # Simple utility
    
    def pop_done(self):
        """
        Pops a done task from the group.
        
        Returns
        -------
        future : `None`, ``Future``
        """
        done = self.done
        if done:
            return done.pop()
    
    
    def has_done(self):
        """
        Returns whether there are done tasks in the group.
        
        Returns
        -------
        has_done : `bool`
        """
        return True if self.done else False
    
    
    def has_pending(self):
        """
        Returns whether there are pending tasks in the group.
        
        Returns
        -------
        has_pending : `bool`
        """
        return True if self.pending else False
    
    
    def cancel_all(self):
        """
        Cancels the done and the pending tasks of the task group.
        """
        for future in self.done:
            future.cancel()
            
        for future in self.pending:
            future.cancel()
    
    
    def cancel_done(self):
        """
        Cancels the done tasks of the task group.
        """
        for future in self.done:
            future.cancel()
    
    
    def cancel_pending(self):
        """
        Cancels the pending tasks of the task group.
        """
        for future in self.pending:
            future.cancel()
    
    # Waiters
    
    def wait_next(self):
        """
        Waits till the next future is completed.
        
        Not like ``.wait_first`` this will not trigger if there are already finished tasks, but will only trigger
        if a pending task is completed.
        
        Returns
        -------
        waiter : ``Future``
        
        Examples
        --------
        ```py
        future_0 = Future(loop)
        future_0.set_result(None)
        future_1 = sleep(2.0)
        
        task_group = TaskGroup(loop, [future_0, future_1])
        
        # After 2 seconds, we should retrieve `future_1`.
        next_done = await task_group.wait_next()
        
        assert next_done is future_1
        ```
        """
        return self._add_handler(_handler_wait_next())
    
    
    def wait_first(self):
        """
        Waits for the first task to complete and propagates it.
        
        If there are already done tasks, propagates one of them.
        
        Returns
        -------
        waiter : ``Future`` -> `None` | ``Future``
            After awaiting the waiter propagates the first done task.
        
        Examples
        --------
        ```py
        future_0 = sleep(1.0)
        future_1 = sleep(2.0)
        
        task_group = TaskGroup(loop, [future_0, future_1])
        
        # After 1 seconds, we should retrieve `future_0`.
        first_done = await task_group.wait_first()
        
        assert first_done is future_0
        ```
        """
        return self._add_handler(_handler_wait_first(self))
    
    
    def wait_first_and_pop(self):
        """
        Waits for the first task to complete and propagates it.
        
        If there are already done tasks, propagates one of them.
        
        Removes the propagated task from the task group.
        
        Returns
        -------
        waiter : ``Future`` -> `None` | ``Future``
            After awaiting the waiter propagates the first done task.
        """
        return self._add_handler(_handler_wait_first_and_pop(self))
    
    
    def wait_exception(self):
        """
        Waits till the first task fails with an exception, or till all is done obviously.
        
        If there is any task in the task group failing with an exception, propagates that.
        
        Returns
        -------
        waiter : ``Future`` -> `None` | ``Future``
            After awaiting the waiter propagates the first task that failed with an exception.

        Examples
        --------
        ```py
        future_0 = Future(loop)
        future_0.apply_timeout(1.0)
        future_1 = sleep(2.0)
        
        task_group = TaskGroup(loop, [future_0, future_1])
        
        # After 1 second, we should retrieve `future_0`.
        first_failing_with_exception = await task_group.wait_exception()
        
        assert first_failing_with_exception is future_0
        ```
        """
        return self._add_handler(_handler_wait_exception(self))
    
    
    def wait_exception_and_pop(self):
        """
        Waits till the first task fails with an exception or till all is done obviously.
        
        If there is any task in the task group failing with an exception, propagates that.
        
        Familiar to ``.wait_exception``, but removes the propagated task from the task group.
        
        Returns
        -------
        waiter : ``Future`` -> `None` | ``Future``
            After awaiting the waiter propagates the first task that failed with an exception.
        """
        return self._add_handler(_handler_wait_exception_and_pop(self))
    
    
    def wait_exception_or_cancellation(self):
        """
        Waits till the first task fails with an exception, is cancelled, or till all is done obviously.
        
        If there is any task in the task group failing with an exception or with cancellation, propagates that.
        
        Returns
        -------
        waiter : ``Future`` -> `None` | ``Future``
            After awaiting the waiter propagates the first task that failed with an exception.

        Examples
        --------
        ```py
        future_0 = sleep(1)
        future_1 = Future(loop)
        future_1.cancel()
        
        task_group = TaskGroup(loop, [future_0, future_1])
        
        # we should retrieve `future_0` as soon as possible.
        first_failing_with_exception_or_cancellation = await task_group.wait_exception()
        
        assert first_failing_with_exception_or_cancellation is future_1
        ```
        """
        return self._add_handler(_handler_wait_exception_or_cancellation(self))
    
    
    def wait_exception_or_cancellation_and_pop(self):
        """
        Waits till the first task fails with an exception, is cancelled or till all is done obviously.
        
        If there is any task in the task group failing with an exception or with cancellation, propagates that.
        
        Familiar to ``.wait_exception``, but removes the propagated task from the task group.
        
        Returns
        -------
        waiter : ``Future`` -> `None` | ``Future``
            After awaiting the waiter propagates the first task that failed with an exception.
        """
        return self._add_handler(_handler_wait_exception_or_cancellation_and_pop(self))
    
    
    def wait_first_n(self, count):
        """
        Waits for the first `n` tasks to complete.
        
        Parameters
        ----------
        count : `int`
            The required amount of futures to wait for.
        
        Returns
        -------
        waiter : ``Future`` -> `int`
            After awaiting the waiter propagates the amount of done tasks.
            
            This amount can be higher than `count` if there are initially more done tasks in the task group.
            It can also be less than the amount of done tasks inside of task group if any was popped from it.
        
        Examples
        --------
        ```py
        future_0 = sleep(0.0)
        future_1 = sleep(1.0)
        future_2 = sleep(2.0)
        
        task_group = TaskGroup(loop, [future_0, future_1, future_2])
        
        # After 1 seconds 2 tasks should be done.
        done_count = await task_group.wait_first_n(2)
        
        assert done_count == 2
        assert future_0.is_done()
        assert future_1.is_done()
        assert future_2.is_pending()
        ```
        """
        return self._add_handler(_handler_wait_first_n(self, count))
    
    
    def wait_all(self):
        """
        Waits till all the tasks are completed.
        
        Returns
        -------
        waiter : ``Future``
            A future to be waited propagating the task group itself.
        
        Examples
        --------
        ```py
        future_0 = sleep(0.0)
        future_1 = sleep(1.0)
        
        task_group = TaskGroup(loop, [future_0, future_1])
        
        # After 2 seconds all tasks are finished.
        await task_group.wait_all()
        
        assert not task_group.has_pending()
        ```
        """
        return self._add_handler(_handler_wait_all(self))
    
    # Iterators
    
    def iter_futures(self):
        """
        Iterates over the tasks of the task group.
        
        This function is an iterable generator.
        
        Yields
        ------
        future : ``Future``
        """
        yield from self.done
        yield from self.pending
    
    
    def exhaust_done(self):
        """
        Iterates over the done tasks ands pops them.
        
        This method is an iterable generator.
        
        Yields
        ------
        future : ``Future``
        
        Examples
        --------
        ```py
        task_group = TaskGroup(loop)
        
        task_group.create_future().set_result(None)
        task_group.create_future().set_result(None)
        task_group.add_future(sleep(1.0))
        
        # We have 2 done futures and 1 pending.
        done = [*task_group.exhaust_done()]
        
        assert len(done) == 2
        assert not task_group.has_done()
        assert task_group.has_pending()
        ```
        """
        done = self.done
        while done:
            yield done.pop()
    
    
    async def exhaust(self):
        """
        Async iterates over the added tasks. As one is done, yields and pops it.
        
        This method is a coroutine generator.
        
        Yields
        ------
        future : ``Future``
        
        Examples
        --------
        ```py
        future_0 = sleep(0.0)
        future_1 = sleep(1.0)
        future_2 = sleep(2.0)
        task_group = TaskGroup(loop, [future_0, future_1, future_2])
        
        # After 2 second our iteration should end with the last done future.
        done_futures = []
        async for future in task_group.exhaust():
            done_futures.append(future)
        
        # We should get back the futures in their completion order.
        assert done_futures = [future_0, future_1, future_2]
        ```
        """
        done = self.done
        pending = self.pending
        
        while True:
            while done:
                yield done.pop()
            
            if not pending:
                return
            
            waiter = self.wait_next()
            try:
                await waiter
            except GeneratorExit:
                waiter.cancel()
                raise
    
    # Context managers
    
    def cancel_on_exception(self):
        """
        Returns a context manager that can be used to cancel the task group if an exception occurs.
        
        Returns
        -------
        context_manager : ``TaskGroupContextManager``
        
        Examples
        --------
        ```py
        task_group = TaskGroup(loop)
        future = task_group.create_future()
        
        try:
            with task_group.cancel_on_exception():
                raise ValueError()
        finally:
            assert future.is_cancelled()
        ```
        """
        return TaskGroupContextManager(self, _context_manager_leaver_cancel_on_exception)



__all__ = ('CallableAnalyzer', 'RichAnalyzer')

import sys
from itertools import zip_longest
from types import FunctionType

from .async_utils import is_coroutine_function, is_coroutine_generator_function
from .code import CO_VARARGS, CO_VARKEYWORDS
from .method_like import MethodLike


IS_PYTHON_STULTUS = sys.version_info >= (3, 10, 0)

INSTANCE_TO_ASYNC_FALSE = 1
INSTANCE_TO_ASYNC_TRUE = 2
INSTANCE_TO_ASYNC_CANNOT = 3
INSTANCE_TO_ASYNC_GENERATOR_FALSE = 4
INSTANCE_TO_ASYNC_GENERATOR_TRUE = 5

PARAMETER_TYPE_POSITIONAL_ONLY = 1
PARAMETER_TYPE_POSITIONAL_AND_KEYWORD = 2
PARAMETER_TYPE_KEYWORD_ONLY = 3
PARAMETER_TYPE_ARGS = 4
PARAMETER_TYPE_KWARGS = 5

PARAMETER_TYPE_NAMES = {
    PARAMETER_TYPE_POSITIONAL_ONLY: 'positional only',
    PARAMETER_TYPE_POSITIONAL_AND_KEYWORD: 'positional',
    PARAMETER_TYPE_KEYWORD_ONLY: 'keyword only',
    PARAMETER_TYPE_ARGS: 'args',
    PARAMETER_TYPE_KWARGS: 'kwargs',
}

class Parameter:
    """
    Represents a callable's parameter.
    
    Attributes
    ----------
    annotation : `Any`
        The parameter's annotation if applicable. Defaults to `None`.
    default : `Any`
        The default value of the parameter if applicable. Defaults to `None`.
    has_annotation : `bool`
        Whether the parameter has annotation.
    has_default : `bool`
        Whether the parameter has default value.
    name : `str`
        The parameter's name.
    positionality : `int`
        Whether the parameter is positional, keyword or such.
        
        Can be set one of the following:
        +----------------------------------------+-----------+
        | Respective Name                        | Value     |
        +========================================+===========+
        | PARAMETER_TYPE_POSITIONAL_ONLY         | 1         |
        +----------------------------------------+-----------+
        | PARAMETER_TYPE_POSITIONAL_AND_KEYWORD  | 2         |
        +----------------------------------------+-----------+
        | PARAMETER_TYPE_KEYWORD_ONLY            | 3         |
        +----------------------------------------+-----------+
        | PARAMETER_TYPE_ARGS                    | 4         |
        +----------------------------------------+-----------+
        | PARAMETER_TYPE_KWARGS                  | 5         |
        +----------------------------------------+-----------+
    reserved : `bool`
        Whether the parameter is reserved.
        
        For example at the case of methods, the first parameter is reserved for the `self` parameter.
    """
    __slots__ = ('annotation', 'default', 'has_annotation', 'has_default', 'name', 'positionality', 'reserved', )
    
    def __repr__(self):
        """Returns the parameter's representation."""
        repr_parts = []
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        repr_parts.append(' ')
        
        if self.reserved:
            repr_parts.append('reserved, ')
        
        repr_parts.append(PARAMETER_TYPE_NAMES[self.positionality])
        
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        if self.has_default:
            repr_parts.append(', default = ')
            repr_parts.append(repr(self.default))
        
        if self.has_annotation:
            repr_parts.append(', annotation = ')
            repr_parts.append(repr(self.annotation))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two parameters are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # has_annotation
        has_annotation = self.has_annotation
        if (has_annotation != other.has_annotation):
            return False
        
        if has_annotation:
            # annotation
            if self.annotation != other.annotation:
                return False
        
        # has_default
        has_default = self.has_default
        if (has_default != other.has_default):
            return False
        
        if has_default:
            # default
            if (self.default is not other.default) or (self.default != other.default):
                return False
        
        # name
        if self.name != other.name:
            return False
        
        # positionality
        if self.positionality != other.positionality:
            return False
        
        # reserved
        if self.reserved != other.reserved:
            return False
        
        return True
    
    
    def is_positional_only(self):
        """
        Returns whether the parameter is positional only.
        
        Returns
        -------
        is_positional_only : `bool`
        """
        positionality = self.positionality
        if positionality == PARAMETER_TYPE_POSITIONAL_ONLY:
            return True
        
        return False
    
    
    def is_positional(self):
        """
        Returns whether the parameter is positional.
        
        Returns
        -------
        is_positional : `bool`
        """
        positionality = self.positionality
        if positionality == PARAMETER_TYPE_POSITIONAL_ONLY:
            return True
        
        if positionality == PARAMETER_TYPE_POSITIONAL_AND_KEYWORD:
            return True
        
        return False
    
    
    def is_keyword(self):
        """
        Returns whether the parameter can be used as a keyword parameter.
        
        Returns
        -------
        is_keyword : `bool`
        """
        positionality = self.positionality
        if positionality == PARAMETER_TYPE_POSITIONAL_AND_KEYWORD:
            return True
        
        if positionality == PARAMETER_TYPE_KEYWORD_ONLY:
            return True
        
        return False
    
    
    def is_keyword_only(self):
        """
        Returns whether they parameter is keyword only.
        
        Returns
        -------
        is_keyword_only : `bool`
        """
        positionality = self.positionality
        if positionality == PARAMETER_TYPE_KEYWORD_ONLY:
            return True
        
        return False
    
    
    def is_args(self):
        """
        Returns whether the parameter is an `*args` parameter.
        
        Returns
        -------
        is_args : `bool`
        """
        positionality = self.positionality
        if positionality == PARAMETER_TYPE_ARGS:
            return True
        
        return False
    
    
    def is_kwargs(self):
        """
        Returns whether the parameter is an `**kwargs` parameter.
        
        Returns
        -------
        is_kwargs : `bool`
        """
        positionality = self.positionality
        if positionality == PARAMETER_TYPE_KWARGS:
            return True
        
        return False

if IS_PYTHON_STULTUS:
    def compile_annotations(real_function, annotations):
        new_annotations = {}
        if not annotations:
            return new_annotations
        
        global_variables = getattr(real_function, '__globals__', None)
        if (global_variables is None):
            # Builtins go brrr
            return new_annotations
        
        for key, value in annotations.items():
            if type(value) is str:
                try:
                    value = eval(value, global_variables, None)
                except:
                    pass
            
            new_annotations[key] = value
        
        return new_annotations


class CallableAnalyzer:
    """
    Analyzer for callable-s.
    
    Can analyze functions, methods, callable objects and types or such.
    
    Attributes
    ----------
    args_parameter : `None`, ``Parameter``
        If the analyzed callable has `*args` parameter, then this attribute is set to it. Defaults to `None`.
    parameters : `list` of ``Parameter``
        The analyzed callable's parameters.
    callable : `callable`
        The analyzed object.
    instance_to_async : `int`
        Whether the analyzed object can be instanced to async.
        
        +---------------------------+-----------+-------------------------------------------+
        | Respective Name           | Value     | Description                               |
        +===========================+===========+===========================================+
        | INSTANCE_TO_ASYNC_FALSE   | 1         | Whether the object is async.              |
        +---------------------------+-----------+-------------------------------------------+
        | INSTANCE_TO_ASYNC_TRUE    | 2         | Whether the object is on async callable,  |
        |                           |           | but after instancing it, returns one.     |
        +---------------------------+-----------+-------------------------------------------+
        | INSTANCE_TO_ASYNC_CANNOT  | 3         | Whether the object is not async.          |
        +---------------------------+-----------+-------------------------------------------+
    kwargs_parameter : `None`, ``Parameter``
        If the analyzed callable has `**kwargs`, then this attribute is set to it. Defaults to `None`.
    method_allocation : `int`
        How much parameter is allocated if the analyzed callable is method if applicable.
    real_function : `callable`
        The function wrapped by the given callable.
    """
    __slots__ = (
        'args_parameter', 'parameters', 'callable', 'instance_to_async', 'kwargs_parameter', 'method_allocation',
        'real_function'
    )
    
    def __repr__(self):
        """Returns the callable analyzer's representation."""
        repr_parts = []
        repr_parts.append('<')
        repr_parts.append(self.__class__.__name__)
        
        if self.is_async():
            repr_parts.append(' async')
        elif self.is_async_generator():
            repr_parts.append(' async generator')
        elif self.can_instance_to_async_callable():
            repr_parts.append(' instance async')
        elif self.can_instance_to_async_generator():
            repr_parts.append(' instance async generator')
        
        method_allocation = self.method_allocation
        if method_allocation:
            repr_parts.append(' method')
            if method_allocation!=1:
                repr_parts.append(' (')
                repr_parts.append(repr(method_allocation))
                repr_parts.append(')')
        
        repr_parts.append(' ')
        callable_ = self.callable
        repr_parts.append(repr(callable_))
        real_function = self.real_function
        if (callable_ is not real_function):
            repr_parts.append(' (')
            repr_parts.append(repr(real_function))
            repr_parts.append(')')
        
        repr_parts.append(', parameters = ')
        repr_parts.append(repr(self.parameters))
        
        args_parameter = self.args_parameter
        if (args_parameter is not None):
            repr_parts.append(', args = ')
            repr_parts.append(repr(args_parameter))
        
        kwargs_parameter = self.kwargs_parameter
        if (kwargs_parameter is not None):
            repr_parts.append(', kwargs = ')
            repr_parts.append(repr(kwargs_parameter))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __new__(cls, callable_, as_method=False):
        """
        Analyzes the given callable.
        
        Parameters
        ----------
        callable_ : `callable`
            The callable to analyze.
        as_method : `bool` = `False`, Optional
            Whether the given `callable` is given as a `function`, but it should be analyzed as a `method`.
        
        Raises
        ------
        TypeError
            If the given object is not callable, or could not be used as probably intended.
        """
        while True:
            if isinstance(callable_, FunctionType):
                
                real_function = callable_
                if is_coroutine_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_FALSE
                elif is_coroutine_generator_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                else:
                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                
                method_allocation = 0
                break
            
            if isinstance(callable_, MethodLike):
                real_function = callable_
                
                if is_coroutine_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_FALSE
                elif is_coroutine_generator_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                else:
                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                
                method_allocation = MethodLike.get_reserved_parameter_count(callable_)
                break
            
            if not isinstance(callable_, type) and hasattr(type(callable_), '__call__'):
                real_function = type(callable_).__call__
                
                if is_coroutine_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_FALSE
                elif is_coroutine_generator_function(real_function):
                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                else:
                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                    
                if type(real_function) is FunctionType:
                    method_allocation = 1
                else:
                    method_allocation = MethodLike.get_reserved_parameter_count(real_function)
                
                break
            
            if not issubclass(callable_, type) and isinstance(callable_, type):
                
                while True:
                    real_function = callable_.__new__
                    if not callable(real_function):
                        raise TypeError(
                            f'`{callable_!r}.__new__` should be callable, got {real_function.__class__.__name__}; '
                            f'{real_function!r}.'
                        )
                    
                    if real_function is not object.__new__:
                        if is_coroutine_function(real_function):
                            instance_to_async = INSTANCE_TO_ASYNC_FALSE
                        elif is_coroutine_generator_function(real_function):
                            instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_FALSE
                        else:
                            if hasattr(callable_, '__call__'):
                                call = callable_.__call__
                                if is_coroutine_function(call):
                                    instance_to_async = INSTANCE_TO_ASYNC_TRUE
                                elif is_coroutine_generator_function(call):
                                    instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_TRUE
                                else:
                                    instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                            else:
                                instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                        
                        if type(real_function) is FunctionType:
                            method_allocation = 1
                        else:
                            method_allocation = MethodLike.get_reserved_parameter_count(real_function)
                        
                        break
                    
                    real_function = callable_.__init__
                    if not callable(real_function):
                        raise TypeError(
                            f'`{callable_!r}.__init__` should be callable, got {type(real_function).__name__}; '
                            f'{real_function!r}.'
                        )
                    
                    if real_function is not object.__init__:
                        if hasattr(callable_, '__call__'):
                            call = callable_.__call__
                            if is_coroutine_function(call):
                                instance_to_async = INSTANCE_TO_ASYNC_TRUE
                            elif is_coroutine_generator_function(call):
                                instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_TRUE
                            else:
                                instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                        else:
                            instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                        
                        if type(real_function) is FunctionType:
                            method_allocation = 1
                        else:
                            method_allocation = MethodLike.get_reserved_parameter_count(real_function)
                        
                        break
                    
                    real_function = None
                    method_allocation = 0
                    
                    if hasattr(callable_, '__call__'):
                        call = callable_.__call__
                        if is_coroutine_function(call):
                            instance_to_async = INSTANCE_TO_ASYNC_TRUE
                        elif is_coroutine_generator_function(call):
                            instance_to_async = INSTANCE_TO_ASYNC_GENERATOR_TRUE
                        else:
                            instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                    else:
                        instance_to_async = INSTANCE_TO_ASYNC_CANNOT
                    break
                
                break
            
            raise TypeError(
                f'Expected `FunctionType`, `MethodType`, `callable`, got {callable_.__class__.__name__}; {callable_!r}.'
            )
        
        if as_method and type(callable_) is FunctionType:
            method_allocation += 1
        
        if (real_function is not None) and not hasattr(real_function, '__code__'):
            raise TypeError(
                f'Expected `function-like`, got {real_function.__class__.__name__}; {real_function!r}.'
            )
        
        parameters = []
        if (real_function is not None):
            parameter_count = real_function.__code__.co_argcount
            accepts_args = real_function.__code__.co_flags & CO_VARARGS
            keyword_only_parameter_count = real_function.__code__.co_kwonlyargcount
            accepts_kwargs = real_function.__code__.co_flags & CO_VARKEYWORDS
            positional_only_parameter_count = getattr(real_function.__code__, 'co_posonlyargcount', 0)
            default_parameter_values = real_function.__defaults__
            default_keyword_only_parameter_values = real_function.__kwdefaults__
            annotations = getattr(real_function, '__annotations__', None)
            if (annotations is None):
                annotations = {}
            elif IS_PYTHON_STULTUS:
                annotations = compile_annotations(real_function, annotations)
            
            start = 0
            end = parameter_count
            parameter_names = real_function.__code__.co_varnames[start:end]
            
            start = end
            end = start + keyword_only_parameter_count
            keyword_only_parameter_names = real_function.__code__.co_varnames[start:end]
            
            if accepts_args:
                args_name = real_function.__code__.co_varnames[end]
                end += 1
            else:
                args_name = None
            
            if accepts_kwargs:
                kwargs_name = real_function.__code__.co_varnames[end]
            else:
                kwargs_name = None
            
            names_to_defaults = {}
            if (default_parameter_values is not None) and default_parameter_values:
                parameter_index = parameter_count - len(default_parameter_values)
                default_index = 0
                while parameter_index < parameter_count:
                    name = parameter_names[parameter_index]
                    default = default_parameter_values[default_index]
                    
                    names_to_defaults[name] = default
                    
                    parameter_index += 1
                    default_index += 1
            
            if (default_keyword_only_parameter_values is not None) and default_keyword_only_parameter_values:
                parameter_index = keyword_only_parameter_count - len(default_keyword_only_parameter_values)
                while parameter_index < keyword_only_parameter_count:
                    name = keyword_only_parameter_names[parameter_index]
                    default = default_keyword_only_parameter_values[name]
                    
                    names_to_defaults[name] = default
                    
                    parameter_index += 1
            
            if (method_allocation > parameter_count) and (args_name is None):
                raise TypeError(
                    f'Received a `method-like`, but has not enough positional parameters, got '
                    f'{real_function.__class__.__name__}; {real_function!r}; '
                    f'allocated parameter count={method_allocation!r}; total parameter count={parameter_count!r}.'
                )
            
            index = 0
            while index < parameter_count:
                parameter = Parameter()
                name = parameter_names[index]
                parameter.name = name
                
                try:
                    annotation = annotations[name]
                except KeyError:
                    parameter.has_annotation = False
                    parameter.annotation = None
                else:
                    parameter.has_annotation = True
                    parameter.annotation = annotation
                
                try:
                    default = names_to_defaults[name]
                except KeyError:
                    parameter.has_default = False
                    parameter.default = None
                else:
                    parameter.has_default = True
                    parameter.default = default
                
                if index < positional_only_parameter_count:
                    parameter.positionality = PARAMETER_TYPE_POSITIONAL_ONLY
                else:
                    parameter.positionality = PARAMETER_TYPE_POSITIONAL_AND_KEYWORD
                
                parameter.reserved = (index<method_allocation)
                parameters.append(parameter)
                index = index + 1
            
            if args_name is None:
                args_parameter = None
            else:
                args_parameter = Parameter()
                args_parameter.name = args_name
                
                try:
                    annotation = annotations[args_name]
                except KeyError:
                    args_parameter.has_annotation = False
                    args_parameter.annotation = None
                else:
                    args_parameter.has_annotation = True
                    args_parameter.annotation = annotation

                args_parameter.has_default = False
                args_parameter.default = None
                args_parameter.positionality = PARAMETER_TYPE_ARGS
                
                if method_allocation > parameter_count:
                    args_parameter.reserved = True
                else:
                    args_parameter.reserved = False
                parameters.append(args_parameter)
            
            index = 0
            while index < keyword_only_parameter_count:
                parameter = Parameter()
                name = keyword_only_parameter_names[index]
                parameter.name = name
                
                try:
                    annotation = annotations[name]
                except KeyError:
                    parameter.has_annotation = False
                    parameter.annotation = None
                else:
                    parameter.has_annotation = True
                    parameter.annotation = annotation
                
                try:
                    default = names_to_defaults[name]
                except KeyError:
                    parameter.has_default = False
                    parameter.default = None
                else:
                    parameter.has_default = True
                    parameter.default = default
                
                parameter.positionality = PARAMETER_TYPE_KEYWORD_ONLY
                parameter.reserved = False
                parameters.append(parameter)
                index = index + 1
            
            if kwargs_name is None:
                kwargs_parameter = None
            else:
                kwargs_parameter = Parameter()
                kwargs_parameter.name = kwargs_name
                try:
                    annotation = annotations[kwargs_name]
                except KeyError:
                    kwargs_parameter.has_annotation = False
                    kwargs_parameter.annotation = None
                else:
                    kwargs_parameter.has_annotation = True
                    kwargs_parameter.annotation = annotation
                
                kwargs_parameter.has_default = False
                kwargs_parameter.default = None
                kwargs_parameter.positionality = PARAMETER_TYPE_KWARGS
                kwargs_parameter.reserved = False
                parameters.append(kwargs_parameter)
        
        else:
            args_parameter = None
            kwargs_parameter = None
        
        self = object.__new__(cls)
        self.parameters = parameters
        self.args_parameter = args_parameter
        self.kwargs_parameter = kwargs_parameter
        self.callable = callable_
        self.method_allocation = method_allocation
        self.real_function = real_function
        self.instance_to_async = instance_to_async
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two analyzers are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.callable != other.callable:
            return False
        
        return True
    
    
    def __mod__(self, other):
        """Returns whether the two analyzers kinda match."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.instance_to_async != other.instance_to_async:
            return False
        
        for self_parameter, other_parameter in zip_longest(
            self.iter_non_reserved_parameters(), other.iter_non_reserved_parameters()
        ):
            if self_parameter != other_parameter:
                return False
        
        return True
    
    
    def is_async(self):
        """
        Returns whether the analyzed callable is async.
        
        Returns
        -------
        is_async : `bool`
        """
        if self.instance_to_async == INSTANCE_TO_ASYNC_FALSE:
            return True
        
        return False
    
    
    def is_async_generator(self):
        """
        Returns whether the analyzed callable is an async generator.
        
        Returns
        is_async_generator : `bool`
        """
        if self.instance_to_async == INSTANCE_TO_ASYNC_GENERATOR_FALSE:
            return True
        
        return False
    
    
    def can_instance_to_async_callable(self):
        """
        Returns whether the analyzed callable can be instanced to async.
        
        Returns
        -------
        can_instance_to_async_callable : `bool`
        """
        if self.instance_to_async != INSTANCE_TO_ASYNC_TRUE:
            return False
        
        for parameter in self.parameters:
            if parameter.reserved:
                continue
            
            if parameter.has_default:
                continue
            
            return False
        
        return True
    
    
    def can_instance_to_async_generator(self):
        """
        Returns whether the analyzed callable can be instanced to async.
        
        Returns
        -------
        can_instance_to_async_callable : `bool`
        """
        if self.instance_to_async != INSTANCE_TO_ASYNC_GENERATOR_TRUE:
            return False
        
        for parameter in self.parameters:
            if parameter.reserved:
                continue
            
            if parameter.has_default:
                continue
            
            return False
        
        return True
    
    
    # call `.can_instance_async_callable`, `.can_instance_to_async_generator` before
    def instance(self):
        """
        Instances the analyzed callable.
        
        Should be called only after a ``.can_instance_async_callable`` call, if it returned `True`.
        
        Returns
        -------
        instance_to_async_callable : `Any`
        """
        return self.callable()
    
    
    def get_non_default_keyword_only_parameter_count(self):
        """
        Returns the amount of non default keyword only parameters of the analyzed callable.
        
        Returns
        -------
        non_default_keyword_only_parameter_count : `int`
        """
        count = 0
        for value in self.parameters:
            if not value.is_keyword_only():
                continue
            
            if value.has_default:
                break
            
            count += 1
            continue
        
        return count
    
    
    def get_non_reserved_keyword_only_parameters(self):
        """
        Returns the non reserved keyword only parameters of the analyzed callable.
        
        Returns
        -------
        non_default_keyword_only_parameters : `int`
        """
        result = []
        for parameter in self.parameters:
            if not parameter.is_keyword_only():
                continue
            
            if parameter.reserved:
                continue
            
            result.append(parameter)
            continue
        
        return result
    
    
    def get_non_reserved_positional_parameters(self):
        """
        Returns the non reserved positional parameters of the analyzed callable.
        
        Returns
        -------
        non_reserved_positional_parameters : `list` of ``Parameter``
        """
        result = []
        for parameter in self.parameters:
            if not parameter.is_positional():
                break
            
            if parameter.reserved:
                continue
            
            result.append(parameter)
            continue
        
        return result
    
    
    def get_non_reserved_positional_parameter_count(self):
        """
        Returns the amount of the non reserved positional parameters of the analyzed callable.
        
        Returns
        -------
        non_reserved_positional_parameters : `int`
        """
        count = 0
        for parameter in self.parameters:
            if not parameter.is_positional():
                break
            
            if parameter.reserved:
                continue
            
            count +=1
            continue
        
        return count
    
    
    def get_non_reserved_non_default_parameter_count(self):
        """
        Returns the amount of the non reserved non default parameters of the analyzed callable.
        
        Returns
        -------
        non_reserved_non_default_parameter_count : `int`
        """
        count = 0
        for parameter in self.parameters:
            if not parameter.is_positional():
                break
            
            if parameter.reserved:
                continue
            
            if parameter.has_default:
                continue
            
            count +=1
            continue
        
        return count
    
    
    def get_non_reserved_positional_parameter_range(self):
        """
        Returns the minimal and the maximal amount how much non reserved positional parameters the analyzed callable
        expects / accepts.
        
        Returns
        -------
        start : `int`
            The minimal amount of non reserved parameters, what the analyzed callable expects.
        end : `int`
            The maximal amount of non reserved parameters, what the analyzed callable accepts.
        
        Notes
        -----
        `*args` parameter is ignored from the calculation.
        """
        iterator = iter(self.parameters)
        start = 0
        for parameter in iterator:
            if not parameter.is_positional():
                return start, start
            
            if parameter.reserved:
                continue
            
            if parameter.has_default:
                break
            
            start += 1
            continue
        
        else:
            return start, start
        
        end = start
        for parameter in iterator:
            if not parameter.is_positional():
                return start, end
            
            if parameter.reserved:
                continue
            
            end += 1
            continue
        
        return start, end
    
    
    def accepts_args(self):
        """
        Returns whether the analyzed callable accepts `*args` parameter.
        
        Returns
        -------
        accepts_args : `bool`
        """
        return (self.args_parameter is not None)
    
    
    def accepts_kwargs(self):
        """
        Returns whether the analyzed callable accepts `**kwargs` parameter.
        
        Returns
        -------
        accepts_kwargs : `bool`
        """
        return (self.kwargs_parameter is not None)
    
    
    def get_parameter(self, parameter_name):
        """
        Returns the parameter for the given name.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        
        Returns
        -------
        parameter : `None, ``Parameter``
        """
        for parameter in self.parameters:
            if parameter.name == parameter_name:
                return parameter
    
    
    def has_parameter(self, parameter_name):
        """
        Returns whether the analyzed callable has the given parameter by name.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        
        Returns
        -------
        has_parameter : `bool`
        """
        for parameter in self.parameters:
            if parameter.name == parameter_name:
                return True
        
        return False
    
    
    def iter_non_reserved_parameters(self):
        """
        Iterates over the non-reserved parameters of the analyzer.
        
        This method is an iterable generator.
        
        Yields
        ------
        parameter : ``Parameter``
        """
        for parameter in self.parameters:
            if not parameter.reserved:
                yield parameter


class RichAnalyzerParameterAccess:
    """
    Parameter access of a ``RichAnalyzer``.
    
    Attributes
    ----------
    _analyzer : ``CallableAnalyzer``
        Analyzer analyzing the respective function.
    """
    def __new__(cls, analyzer):
        self = object.__new__(cls)
        self._analyzer = analyzer
        return self
    
    def __getattr__(self, attribute_name):
        """
        Tries to find the specified attribute of the respective function.
        
        Returns
        -------
        parameter : ``Parameter``
        
        Raises
        ------
        AttributeError
            - If the parameter by the specified name is not found.
        """
        for parameter in self._analyzer.parameters:
            if parameter.name == attribute_name:
                return parameter
        
        raise AttributeError(attribute_name)


class RichAnalyzer:
    """
    Analyzer supporting rich access.
    
    Attributes
    ----------
    _analyzer : ``CallableAnalyzer``
        Analyzer analyzing the respective function.
    """
    def __new__(cls, callable_, as_method=False):
        """
        Analyzes the given callable.
        
        Parameters
        ----------
        callable_ : `callable`
            The callable to analyze.
        as_method : `bool` = `False`, Optional
            Whether the given `callable` is given as a `function`, but it should be analyzed as a `method`. Defaults
            to `False`.
        
        Raises
        ------
        TypeError
            If the given object is not callable, or could not be used as probably intended.
        """
        analyzer = CallableAnalyzer(callable_, as_method=as_method)
        
        self = object.__new__(cls)
        self._analyzer = analyzer
        return self
    
    
    @property
    def name(self):
        """
        Returns the name of the analyzed callable.
        
        Returns
        -------
        name : `str`
        """
        return self._analyzer.__name__
    
    
    @property
    def parameters(self):
        """
        Returns parameter access to the
        Returns
        -------
        parameter_access : RichAnalyzerParameterAccess
        """
        return RichAnalyzerParameterAccess(self._analyzer)



__all__ = ('CancelledError', 'InvalidStateError',)

from ..utils import include


get_future_state_name = include('get_future_state_name')


class CancelledError(BaseException):
    """
    The future or task was cancelled.
    """
    __slots__ = ()


def _create_invalid_state_error_message(error):
    """
    Helper function to create the error message for invalid state errors.
    
    Parameters
    ----------
    error : ``InvalidStateError``
        The invalid state error to create message for.
    
    Returns
    -------
    message : `str`
    """
    future = error.future
    location = error.location
    return (
        f'`{future.__class__.__name__}.{location}` was called, when `._state` is {future._state} '
        f'({get_future_state_name(future._state)}) of {future!r}.'
    )


class InvalidStateError(Exception):
    """
    The operation is not allowed in this state.
    
    Attributes
    ----------
    _message : `None`, `str`
        Internal cache for the ``.message`` property.
    
    _message_given : `bool`
        Whether message was given when constructing.
    
    future : ``Future``.
        The future, from what's method the exception was raised.
    
    location : `str`
        The future's function's name, from where the exception was raised.
    """
    __slots__ = ('_message', '_message_given', 'future', 'location')
    
    # This is to support keyword parameters
    __init__ = object.__init__
    
    def __new__(cls, future, location, message = None):
        """
        Creates a new ``InvalidStateError``.
        
        Parameters
        ----------
        future : ``Future``.
            The future, from what's method the exception was raised.
        location : `str`
            The future's function's name, from where the exception was raised.
        message : `str`
            The exception's message. If not defined, then the exception will generate it by itself.
        """
        self = Exception.__new__(cls, future, location, message)
        self.future = future
        self.location = location
        self._message = message
        self._message_given = message is not None
        return self
    
    
    def __repr__(self):
        """Returns the exception's representation."""
        return f'{self.__class__.__name__}: {self.message}'
    
    
    def __str__(self):
        """Returns the exception's message."""
        return self.message
    
    
    @property
    def message(self):
        """
        Returns the exception's message.
        
        Returns
        -------
        message : `str`
        
        Notes
        -----
        If the exception was created wia ``.__init__``, without giving defining the `message` parameter, then the
        exception's message is generated only when this property is retrieved for the first time.
        """
        message = self._message
        if message is None:
            message = _create_invalid_state_error_message(self)
            self._message = message
        
        return message
    
    
    def __eq__(self, other):
        """Returns whether the two exceptions are equal."""
        if type(other) is not type(self):
            return NotImplemented
        
        if self.future is not other.future:
            return False
        
        if self.location != other.location:
            return False
        
        message_given = self._message_given
        if message_given != other._message_given:
            return False
        
        if message_given and self._message != other._message:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the exception's hash value."""
        hash_value = 0
        
        # future
        hash_value ^= hash(self.future)
        
        # location
        hash_value ^= hash(self.location)
        
        # message
        if self._message_given:
            hash_value ^= hash(self._message)
        
        return hash_value



__all__ = ('RichAttributeErrorBaseType',)

from difflib import get_close_matches
from itertools import permutations as permutate

from .trace import ignore_frame


ignore_frame(__spec__.origin, '__getattr__', 'raise AttributeError(attribute_name)',)
ignore_frame(__spec__.origin, '__getattr__', 'raise AttributeError(exception_message)',)
ignore_frame(__spec__.origin, '__getattr__', 'return object.__getattribute__(self, attribute_name)',)


EXCEPTION_MESSAGE_CACHE = {}

LOCKS = set()

# This will avoid recursion error when something goes really bad.
LOCK_MAX_DEEPNESS = 10

def _iterate_alternative_attribute_names(attribute_name):
    """
    Tries to create alternative versions of the given `attribute_name`.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    attribute_name : `str`
        The attribute name to create alternative names of.
    
    Yields
    ------
    alternative_attribute_name : `str`
    """
    yield attribute_name
    
    split = attribute_name.split('_')
    split_count = len(split)
    if split_count > 1:
        for permutation in permutate(split, split_count):
            yield '_'.join(permutation)


def _get_familiar_attribute_names(instance, attribute_name):
    """
    Gets all the familiar attribute names of the instance to the given one.
    
    Parameters
    ----------
    instance : `object`
        The object to check.
    
    attribute_name : `str`
        The attribute's name to check.
    
    Returns
    -------
    sorted_matches : `None`, `list` of `str`
        Every familiar attribute name. `None` if `attribute_name` is an actual attribute of `instance`.
    """
    directory = dir(instance)
    if (attribute_name in directory):
        return None
    
    collected_matches = set()
    
    diversity = 0.8 - (20 - min(len(attribute_name), 20)) * 0.01
    
    for alternative_attribute_name in _iterate_alternative_attribute_names(attribute_name):
        matches = get_close_matches(alternative_attribute_name, directory, n = 10, cutoff = diversity)
        collected_matches.update(matches)
    
    return sorted(collected_matches)


def _create_rich_exception_message(instance, attribute_name):
    """
    Creates rich exception message for the given instance's given attribute's name.
    
    Parameters
    ----------
    instance : `object`
        The instance who has no attribute like `attribute_name`.
    
    attribute_name : `str`
        The evil variable name.
    
    Returns
    -------
    exception_message : `None`, `str`
        Returns `None` if `attribute_name` is an actual attribute of `instance`.
    """
    lock_key = (type(instance), attribute_name)
    
    if (lock_key not in LOCKS) and (len(LOCKS) < LOCK_MAX_DEEPNESS):
        try:
            LOCKS.add(lock_key)
            
            sorted_matches = _get_familiar_attribute_names(instance, attribute_name)
            if (sorted_matches is None):
                return None
        finally:
            LOCKS.discard(lock_key)
    
    else:
        sorted_matches = None
    
    
    return _build_exception_message(instance, attribute_name, sorted_matches)
    

def _build_exception_message(instance, attribute_name, sorted_matches):
    """
    Builds exception message.
    
    Parameters
    ----------
    instance : `object`
        The instance who has no attribute like `attribute_name`.
    
    attribute_name : `str`
        The evil variable name.
    
    sorted_matches : `None`, `list` of `str`
        Matched attribute names.
    
    Returns
    -------
    exception_message : `str`
    """
    exception_message_parts = []
    
    exception_message_parts.append('`')
    exception_message_parts.append(type(instance).__name__)
    exception_message_parts.append('` has no attribute `')
    exception_message_parts.append(attribute_name)
    exception_message_parts.append('`')
    
    if (sorted_matches is not None):
        matches_count = len(sorted_matches)
        if matches_count == 0:
            pass
        
        elif matches_count == 1:
            exception_message_parts.append('; Did you mean: `')
            exception_message_parts.append(sorted_matches[0])
            exception_message_parts.append('`')
        
        else:
            exception_message_parts.append('; Did you mean any of: ')
            
            index = 0
            while True:
                exception_message_parts.append('`')
                exception_message_parts.append(sorted_matches[index])
                exception_message_parts.append('`')
                
                index += 1
                if index == matches_count:
                    break
                
                exception_message_parts.append(', ')
                continue
    
    exception_message_parts.append('.')
    
    return ''.join(exception_message_parts)


class RichAttributeErrorBaseType:
    """
    Base type for generating rich attribute error messages.
    """
    __slots__ = ()
    
    def __getattr__(self, attribute_name):
        """Drops a rich attribute error."""
        
        # Python tries to get `__dict__` even if there is no `__dict__` when calling `dir()`.
        if (attribute_name == '__dict__') and ((type(self), attribute_name) in LOCKS):
            raise AttributeError(attribute_name)
        
        # Use goto to avoid putting `KeyError` on exception traceback context.
        while True:
            if type(self).__getattr__ is not RichAttributeErrorBaseType.__getattr__:
                exception_message = _create_rich_exception_message(self, attribute_name)
                break
            
            key = (getattr(type(self), '__module__', ''), attribute_name)
            
            try:
                exception_message = EXCEPTION_MESSAGE_CACHE[key]
            except KeyError:
                pass
            else:
                break
            
            exception_message = _create_rich_exception_message(self, attribute_name)
            if (exception_message is not None):
                EXCEPTION_MESSAGE_CACHE[key] = exception_message
            break
        
        if exception_message is None:
            return object.__getattribute__(self, attribute_name)
        
        raise AttributeError(exception_message)



__all__ = ('Future',)

import sys
from datetime import datetime as DateTime
from types import MethodType
from warnings import warn

from ...utils import ignore_frame, include, to_coroutine

from ..exceptions import CancelledError, InvalidStateError

from .future_repr import render_callbacks_into, render_result_into, render_state_into
from .future_states import (
    FUTURE_STATE_CANCELLED, FUTURE_STATE_CANCELLING_SELF, FUTURE_STATE_MASK_DONE, FUTURE_STATE_MASK_SILENCED,
    FUTURE_STATE_RESULT_RAISE, FUTURE_STATE_RESULT_RAISE_RETRIEVED, FUTURE_STATE_RESULT_RETURN, FUTURE_STATE_SILENCED
)


FutureWrapperSync = include('FutureWrapperSync')
FutureWrapperAsync = include('FutureWrapperAsync')
write_exception_async = include('write_exception_async')
write_exception_maybe_async = include('write_exception_maybe_async')

ignore_frame(__spec__.origin, 'get_result', 'raise self._result')
ignore_frame(__spec__.origin, 'get_result', 'raise CancelledError')
ignore_frame(__spec__.origin, '__iter__', 'yield self',)
ignore_frame(__spec__.origin, '__iter__', 'return self.get_result()',)


SILENCE_DEPRECATED =  DateTime.utcnow() > DateTime(2023, 11, 12)


def _set_timeout_if_pending(future):
    """
    Sets timeout exception into the future if still pending.
    
    Parameters
    ----------
    future : ``Future``
        The future to set timeout to.
    """
    future.cancel_with(TimeoutError)


def _cancel_handle_callback(handle, future):
    """
    Callback added to the future to cancel a handle.
    
    Parameters
    ----------
    handle : ``Handle``
        The handle to cancel.
    future : ``Future``
        The completed future.
    """
    handle.cancel()


class Future:
    """
    A Future represents an eventual result of an asynchronous operation.
    
    Future is an awaitable object. Coroutines can await on ``Future`` objects until they either have a result or an
    exception set, or until they are cancelled.
    
    Attributes
    ----------
    _blocking : `bool`
        Whether the future is already being awaited, so it blocks the respective coroutine.
    
    _callbacks : `list` of `callable`
        The callbacks of the future, which are queued up on the respective event loop to be called, when the future is
        finished. These callback should accept `1` parameter, the future itself.
        
        Note, if the future is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
    
    _loop : ``EventThread``
        The loop to what the created future is bound.
    
    _result : `None`, `object`
        The result of the future. Defaults to `None`.
    
    _state : `int`
        The state of the future stored as a bitwise flags.
    """
    __slots__ = ('_blocking', '_callbacks', '_loop', '_result', '_state')
    
    def __new__(cls, loop):
        """
        Creates a new ``Future`` object bound to the given `loop`.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop to what the created future will be bound to.
        """
        self = object.__new__(cls)
        self._blocking = False
        self._callbacks = []
        self._loop = loop
        self._result = None
        self._state = 0
        return self
    
    
    def __repr__(self):
        """Returns the future's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        state = self._state
        repr_parts, field_added = render_state_into(repr_parts, False, state)
        repr_parts, field_added = render_result_into(repr_parts, field_added, state, self._result)
        repr_parts, field_added = render_callbacks_into(repr_parts, field_added, self._callbacks)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def cancel(self):
        """
        Cancels the future if it is pending.
        
        Returns
        -------
        result : `int` (`0`, `1`)
            If the future is already done, returns `0`, if it got cancelled, returns `1`.
        """
        state = self._state
        
        if state & FUTURE_STATE_MASK_DONE:
            state |= FUTURE_STATE_SILENCED
            result = 0
        else:
            state |= FUTURE_STATE_CANCELLED
            result = 1
            self._loop._schedule_callbacks(self)
        
        self._state = state
        return result
    
    
    def cancel_with(self, exception):
        """
        Cancels the future with a custom exception.
        
        Returns
        -------
        result : `int` (`0`, `1`)
            If the future is already done, returns `0`, if it got cancelled, returns `1`.
        """
        state = self._state
        
        if state & FUTURE_STATE_MASK_DONE:
            state |= FUTURE_STATE_SILENCED
            result = 0
        
        else:
            if isinstance(exception, type):
                exception = exception()
            
            if isinstance(exception, StopIteration):
                raise TypeError(
                    f'{exception!r} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
                )
            
            state |= FUTURE_STATE_CANCELLED | FUTURE_STATE_RESULT_RAISE
            self._result = exception
            self._loop._schedule_callbacks(self)
            result = 1
        
        self._state = state
        return result
        
    
    def silence(self):
        """
        Silences cleanup warnings of the future.
        
        When a task is cancelled these warnings are also silenced.
        """
        self._state |= FUTURE_STATE_SILENCED
    
    
    def is_silenced(self):
        """
        Returns whether clean up warning are silenced of the future.
        
        Also returns `True` if it is in a state without possible cleanup warning.
        
        Returns
        -------
        is_silenced : `bool`
        """
        return True if self._state & FUTURE_STATE_MASK_SILENCED else False
    
    
    def is_cancelling(self):
        """
        Returns whether the future is currently being cancelled.
        
        Returns
        -------
        is_cancelling : `bool`
        """
        state = self._state
        # Cancelling flag might not cleared when the future is already done.
        if state & FUTURE_STATE_MASK_DONE:
            return False
        
        return True if state & FUTURE_STATE_CANCELLING_SELF else False
    
    
    def is_cancelled(self):
        """
        Returns whether the future is cancelled.
        
        Returns
        -------
        is_cancelled : `bool`
        """
        return True if self._state & FUTURE_STATE_CANCELLED else False
    
    
    def is_done(self):
        """
        Returns whether the future is done.
        
        Returns
        -------
        is_done : `bool`
        """
        return True if self._state & FUTURE_STATE_MASK_DONE else False
    
        
    def is_pending(self):
        """
        Returns whether the future is pending.
        
        Returns
        -------
        pending : `bool`
        """
        return False if self._state & FUTURE_STATE_MASK_DONE else True
    
    
    def get_result(self):
        """
        Returns the result of the future.
        
        If the future is cancelled, raises ``CancelledError``.
        
        If the future has exception set with `.set_exception`, `.set_exception_if_pending` successfully, then raises
        the given exception.
        
        If the future has result set with `.set_result`, `.set_result_if_pending` successfully, then returns the
        given object.
        
        If the future is not done yet, or is destroyed, raises ``InvalidStateError``.
        
        Returns
        -------
        result : `object`
        
        Raises
        ------
        CancelledError
            The future is cancelled.
        InvalidStateError
            The futures is not done yet.
        TypeError
            The future has non `BaseException` set as exception.
        BaseException
            The future's set exception.
        """
        state = self._state
        
        if state & FUTURE_STATE_RESULT_RETURN:
            return self._result
        
        if state & FUTURE_STATE_RESULT_RAISE:
            self._state = state | FUTURE_STATE_RESULT_RAISE_RETRIEVED
            raise self._result
        
        if state & FUTURE_STATE_CANCELLED:
            raise CancelledError
        
        # still pending or destroyed
        raise InvalidStateError(self, 'get_result')
    
    
    def get_exception(self):
        """
        Returns the future's exception.
        
        If the future is done, returns it's exception. (Cab be `None`)
        
        If the future is cancelled, returns ``CancelledError``.
        
        If the future is not done yet, or is destroyed, raises ``InvalidStateError``.
        
        Returns
        -------
        exception : `None`, `BaseException`
        
        Raises
        ------
        InvalidStateError
            The futures is not done yet.
        """
        state = self._state
        
        if state & FUTURE_STATE_RESULT_RETURN:
            return None
        
        if state & FUTURE_STATE_RESULT_RAISE:
            self._state = state | FUTURE_STATE_RESULT_RAISE_RETRIEVED
            return self._result
        
        if state & FUTURE_STATE_CANCELLED:
            return CancelledError()
        
        # still pending
        raise InvalidStateError(self, 'get_exception')
    
    
    def get_cancellation_exception(self):
        """
        Returns the exception with what the future was cancelled with. This is usually ``CancelledError``.
        
        Returns `None` if it was not yet cancelled.
        
        Returns
        -------
        exception : `None`, `BaseException`
        """
        if not (self._state & FUTURE_STATE_CANCELLED):
            return None
        
        result = self._result
        if (result is not None):
            return result
        
        return CancelledError()
    
    
    def add_done_callback(self, func):
        """
        Adds the given `func` as a callback of the future.
        
        Parameters
        ----------
        `func` : `callable`
            A callback, what is queued up on the respective event loop, when the future is done. These callback should
            accept `1` parameter, the future itself.
            
        Notes
        -----
        If the future is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
        """
        if self._state & FUTURE_STATE_MASK_DONE:
            self._loop.call_soon(func, self)
        else:
            self._callbacks.append(func)
    
    
    def remove_done_callback(self, func):
        """
        Removes the given `func` from the future's ``._callbacks``.
        
        Parameters
        ----------
        func : `object`
            The callback to remove.
        
        Returns
        -------
        count : `int`
            The total count of the removed callbacks.
        """
        callbacks = self._callbacks
        count = 0
        index = len(callbacks)
        while index:
            index -= 1
            if callbacks[index] is func:
                del callbacks[index]
                count += 1
        
        return count
    
    
    def set_result(self, result):
        """
        Marks the future as done and set's it's result.
        
        Parameters
        ----------
        result : `object`
            The object to set as result.
        
        Raises
        ------
        InvalidStateError
            If the future is already done.
        """
        state = self._state
        if state & FUTURE_STATE_MASK_DONE:
            raise InvalidStateError(self, 'set_result')
            
        self._result = result
        self._state = state | FUTURE_STATE_RESULT_RETURN
        self._loop._schedule_callbacks(self)
    
    
    def set_result_if_pending(self, result):
        """
        Marks the future as done and set's it's result. Not like ``.set_result``, this method will not raise
        ``InvalidStateError`` if the future is already done.
        
        Parameters
        ----------
        result : `object`
            The object to set as result.
        
        Returns
        ------
        set_result : `int` (`0`, `1`)
            If the future is already done, returns `0`, else `1`.
        """
        state = self._state
        if state & FUTURE_STATE_MASK_DONE:
            return 0
        
        self._result = result
        self._state = state | FUTURE_STATE_RESULT_RETURN
        self._loop._schedule_callbacks(self)
        return 1
    
    
    def set_exception(self, exception):
        """
        Marks the future as done and set's it's exception.
        
        Parameters
        ----------
        exception : `BaseException`, `type<BaseException>`
            The exception to set as the future's exception.
        
        Raises
        ------
        InvalidStateError
            If the future is already done.
        TypeError
            If `StopIteration` is given as `exception`.
        """
        state = self._state
        if state & FUTURE_STATE_MASK_DONE:
            raise InvalidStateError(self, 'set_exception')
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception!r} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        self._result = exception
        self._state = state | FUTURE_STATE_RESULT_RAISE
        self._loop._schedule_callbacks(self)
    
    
    def set_exception_if_pending(self, exception):
        """
        Marks the future as done and set's it's exception. Not like ``.set_exception``, this method will not raise
        ``InvalidStateError`` if the future is already done.
        
        Parameters
        ----------
        exception : `BaseException`, `type<BaseException>`
            The exception to set as the future's exception.
        
        Returns
        ------
        set_result : `int` (`0`, `1`)
            If the future is already done, returns `0`, else `1`.
        
        Raises
        ------
        TypeError
            If `StopIteration` is given as `exception`.
        """
        state = self._state
        if state & FUTURE_STATE_MASK_DONE:
            return 0
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception!r} cannot be raised to a(n) {self.__class__.__name__}; {self!r}.'
            )
        
        self._result = exception
        self._state = state | FUTURE_STATE_RESULT_RAISE
        self._loop._schedule_callbacks(self)
        return 1
    
    
    def __iter__(self):
        """
        Awaits the future till it is done.
        
        This method is a generator. Should be used with `await` expression.
        """
        if not (self._state & FUTURE_STATE_MASK_DONE):
            self._blocking = True
            yield self
        
        return self.get_result()
    
    __await__ = __iter__
    
    
    @to_coroutine
    def wait_for_completion(self):
        """
        Awaits the future till it is done, not retrieving it's result.
        
        This method is an awaitable generator.
        
        Notes
        -----
        This method do not protects the future from task cancellation, or from timeout context managers.
        """
        if not (self._state & FUTURE_STATE_MASK_DONE):
            self._blocking = True
            yield self
    
    
    def __del__(self):
        """
        If the future is pending, but it's result was not set, meanwhile anything awaits at it, notifies it.
        
        Also notifies if the future's result was set with ``.set_exception``, or with
        ``.set_exception_if_pending``, and it was not retrieved.
        """
        if not self._loop.running:
            return
        
        state = self._state
        # Do not notify if we have any silenced state
        if state & FUTURE_STATE_MASK_SILENCED:
            return
        
        # If we are not done, btu have callbacks, we want to notify
        if not (state & FUTURE_STATE_MASK_DONE):
            if self._callbacks:
                sys.stderr.write(
                    f'{self.__class__.__name__} is not finished, but still pending: {self!r}\n',
                )
                sys.stderr.flush()
            
            return
        
        # Notify un-retrieved exceptions
        if state & FUTURE_STATE_RESULT_RAISE:
            write_exception_maybe_async(
                self._result,
                f'{self.__class__.__name__} exception was never retrieved: {self!r}\n',
            )
            return
    
    
    def sync_wrap(self):
        """
        Wraps the future, so it's result can be retrieved from a sync thread.
        
        Returns
        -------
        future_wrapper : ``FutureWrapperSync``
            A future awaitable from sync threads.
        """
        return FutureWrapperSync(self)
    
    
    def async_wrap(self, loop):
        """
        Wraps the future, so it's result can be retrieved from an other async thread.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop, from where the future would be awaited.
        
        Returns
        -------
        future_wrapper : ``FutureWrapperAsync``
            An awaitable future from the given event loop.
        """
        return FutureWrapperAsync(self, loop)
    
    
    def apply_timeout(self, timeout):
        """
        Applies timeout to the future. After the timeout duration passed on seconds propagates a `TimeoutError` if the
        future is still pending.
        
        timeout : `float`
            The time after the given `future`'s exception is set as `TimeoutError`.
        
        Returns
        -------
        timeout_applied : `int`
            Returns `1` if timeout was applied. `0` if the future is already finished.
        """
        if timeout <= 0.0:
            return self.cancel_with(TimeoutError)
        
        else:
            if self.is_done():
                return 0
            
            handle = self._loop.call_after(timeout, _set_timeout_if_pending, self)
            if (handle is not None):
                self.add_done_callback(MethodType(_cancel_handle_callback, handle))
        
        return 1
    
    
    def iter_callbacks(self):
        """
        Iterates over the future's callbacks.
        
        This method is an iterable generator.
        
        Yields
        ------
        callback : `object`
        """
        callbacks = self._callbacks
        if (callbacks is not None):
            yield from callbacks
    
    
    # Deprecations
    
    
    def cancelled(self):
        warn(
            (
                f'`{self.__class__.__name__}.cancelled` is deprecated and will be removed in 2023 November. '
                f'Please use `.is_cancelled()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_cancelled()
    
    
    def pending(self):
        warn(
            (
                f'`{self.__class__.__name__}.pending` is deprecated and will be removed in 2023 November. '
                f'Please use `.is_pending()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_pending()
    
    
    def done(self):
        warn(
            (
                f'`{self.__class__.__name__}.done` is deprecated and will be removed in 2023 November. '
                f'Please use `.is_done()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_done()
    
    
    @property
    def result(self):
        warn(
            (
                f'`{self.__class__.__name__}.result` is deprecated and will be removed in 2023 November. '
                f'Use `.get_result()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.get_result
    
    
    @property
    def exception(self):
        warn(
            (
                f'`{self.__class__.__name__}.exception` is deprecated and will be removed in 2023 November. '
                f'Use `.get_exception()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.get_exception
    
    
    if __debug__:
        def __silence__(self):
            """
            Silences the future's `__del__`, so it will not notify if it would.
            
            Deprecated and will be removed in 2024 February.
            
            Notes
            -----
            This method is present only if `__debug__` is set as `True`.
            """
            if SILENCE_DEPRECATED:
                warn(
                    (
                        f'`{self.__class__.__name__}.__silence__` is deprecated and will be removed in 2024 february.'
                        f'Please use `.silence()` instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            self.silence()
    
    
    @property
    def _exception(self):
        """
        Deprecated and will be removed in 2024 february.
        """
        warn(
            f'`{self.__class__.__name__}._exception` is deprecated and will be removed in 2024 february.',
            FutureWarning,
            stacklevel = 2,
        )
        
        if self._state & FUTURE_STATE_RESULT_RAISE:
            return self._result



__all__ = ('get_frame_module', 'get_last_module_frame',)

import sys
from importlib.util import module_from_spec
from sys import _getframe as get_frame


MODULE_FRAME_NAME = '<module>'

def get_last_module_frame():
    """
    Gets the last module frame.
    
    Returns
    -------
    frame : `None`, `FrameType`
    """
    frame = get_frame()
    
    while True:
        frame = frame.f_back
        if frame is None:
            break
        
        # They might falsely name frames with auto generated functions, so check variables as well
        if (frame.f_code.co_name != MODULE_FRAME_NAME):
            continue
        
        # Try to not compare locals with globals if we are inside of a function.
        if (frame.f_locals is not frame.f_globals):
            continue
        
        break
    
    return frame


def get_frame_module(frame):
    """
    Gets from which module the frame is from.
    
    Parameters
    ----------
    frame : `None`, `FrameType`
        The frame to get it's module of.
    
    Returns
    -------
    module : `None`, `ModuleType`
    """
    if frame is None:
        return
    
    spec = frame.f_globals.get('__spec__', None)
    if spec is not None:
        try:
            return sys.modules[spec.name]
        except KeyError:
            pass
        
        return module_from_spec(spec)
    
    system_parameters = sys.argv
    if len(system_parameters) < 1:
        return
    
    executed_file_name = system_parameters[0]
    
    if frame.f_code.co_filename == executed_file_name:
        return sys.modules.get('__main__', None)



from .async_executing import *
from .event import *
from .future import *
from .future_chaining import *
from .future_states import *
from .future_wrapper_async import *
from .future_wrapper_base import *
from .future_wrapper_sync import *
from .handle_cancellers import *
from .locks import *
from .queues import *
from .task import *
from .task_group import *
from .task_suppression import *
from .task_thread_switcher import *
from .timeouting import *


__all__ = (
    'FutureSyncWrapper',
    'FutureAsyncWrapper',
    *async_executing.__all__,
    *event.__all__,
    *future.__all__,
    *future_chaining.__all__,
    *future_states.__all__,
    *future_wrapper_async.__all__,
    *future_wrapper_base.__all__,
    *future_wrapper_sync.__all__,
    *handle_cancellers.__all__,
    *locks.__all__,
    *queues.__all__,
    *task.__all__,
    *task_group.__all__,
    *task_suppression.__all__,
    *task_thread_switcher.__all__,
    *timeouting.__all__,
)

# Deprecations

from warnings import warn


class FutureSyncWrapper(FutureWrapperSync):
    """
    Deprecated and will be removed in 2024 february. Please use ``FutureWrapperSync`` instead.
    """
    __slots__ = ()
        
    def __new__(cls, future, loop):
        warn(
            (
                f'`{cls.__name__}` is deprecated and will be removed in 2024. '
                f'Please use {FutureWrapperSync.__name__} instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return FutureWrapperSync.__new__(cls, future, loop)


class FutureAsyncWrapper(FutureWrapperAsync):
    """
    Deprecated and will be removed in 2024 february. Please use ``FutureWrapperAsync`` instead.
    """
    __slots__ = ()
    
    def __new__(cls, future, loop):
        warn(
            (
                f'`{cls.__name__}` is deprecated and will be removed in 2024. '
                f'Please use {FutureWrapperAsync.__name__} instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return FutureWrapperAsync.__new__(cls, future, loop)



__all__ = ()

from ...utils import export


FUTURE_STATE_RESULT_RETURN = 1 << 0
FUTURE_STATE_RESULT_RAISE = 1 << 1
FUTURE_STATE_RESULT_RAISE_RETRIEVED = 1 << 2

FUTURE_STATE_CANCELLING_SELF = 1 << 3
FUTURE_STATE_CANCELLED = 1 << 4

FUTURE_STATE_DESTROYED = 1 << 5
FUTURE_STATE_SILENCED = 1 << 6


FUTURE_STATE_MASK_RESULT = (
    FUTURE_STATE_RESULT_RETURN |
    FUTURE_STATE_RESULT_RAISE
)

FUTURE_STATE_MASK_DONE = (
    FUTURE_STATE_RESULT_RETURN |
    FUTURE_STATE_RESULT_RAISE |
    FUTURE_STATE_CANCELLED |
    FUTURE_STATE_DESTROYED
)

FUTURE_STATE_MASK_SILENCED = (
    FUTURE_STATE_RESULT_RETURN |
    FUTURE_STATE_RESULT_RAISE_RETRIEVED |
    FUTURE_STATE_CANCELLED |
    FUTURE_STATE_DESTROYED |
    FUTURE_STATE_SILENCED
)


DONE_MASK_AND_NAME_PAIRS = (
    (FUTURE_STATE_RESULT_RETURN, 'result~return'),
    (FUTURE_STATE_RESULT_RAISE_RETRIEVED, 'result~raise~retrieved'),
    (FUTURE_STATE_RESULT_RAISE, 'result~raise'),
    (FUTURE_STATE_DESTROYED, 'destroyed'),
    (FUTURE_STATE_CANCELLED, 'cancelled'),
)


@export
def get_future_state_name(state):
    """
    Gets the future's state's name.
    
    A future can have multiple states at once. They are separated with a comma.
    
    Parameters
    ----------
    state : `int`
        Future state flags.
    
    Returns
    -------
    state_name : `str`
    """
    return ''.join(render_future_state_name_into([], state))


def _add_first_state_name_from(into, field_added, state, mask_name_pairs):
    """
    Adds the first state name of the given ones that matches.
    
    Parameters
    ----------
    into : `list<str>`
        List to render the state into.
    field_added : `bool`
        Whether there were a field added already.
    state : `int`
        Future state flags.
    mask_name_pairs : `tuple<(int, str)>
        Mask and name pairs.
    
    Returns
    -------
    into : `list<str>`
    field_added : `bool`
    """
    for mask, name in mask_name_pairs:
        if state & mask:
            _add_state_name(into, field_added, name)
            field_added = True
            break
    
    return into, field_added


def _add_state_name(into, field_added, name):
    """
    Adds a single state name.
    
    Parameters
    ----------
    into : `list<str>`
        List to render the state into.
    field_added : `bool`
        Whether there were a field added already.
    name : `str`
        the state's name.

    Returns
    -------
    into : `list<str>`
    """
    if field_added:
        into.append(', ')
    into.append(name)
    return into


def render_future_state_name_into(into, state):
    """
    Renders the future's state into the given container.
    
    A future can have multiple states at once. They are separated with a comma.
    
    Parameters
    ----------
    into : `list<str>`
        List to render the state into.
    state : `int`
        Future state flags.
    
    Returns
    -------
    into : `list<str>`
        List to render the state into.
    """
    field_added = False
    
    if state & FUTURE_STATE_MASK_DONE:
        into, field_added = _add_first_state_name_from(into, field_added, state, DONE_MASK_AND_NAME_PAIRS)
    else:
        into = _add_state_name(into, field_added, 'pending')
        field_added = True
        
        if state & FUTURE_STATE_CANCELLING_SELF and not (state & FUTURE_STATE_CANCELLED):
            into = _add_state_name(into, field_added, 'cancelling~self')
            field_added = True
    
    if state & FUTURE_STATE_SILENCED:
        into = _add_state_name(into, field_added, 'silenced')
        
    return into



__all__ = ('FutureWrapperAsync', )

from types import MethodType

from ...utils import export

from .future import Future
from .future_wrapper_base import CANCELLATION_TIMEOUT, FutureWrapperBase


def _set_future_waiter_callback(waiter, future):
    """
    Sets waiter result.
    
    Parameters
    ----------
    waiter : ``Future``
        Waiter to set.
    future : ``Future``
        The finished future.
    """
    if waiter.set_result_if_pending(True):
        waiter._loop.wake_up()


def _skip_loop_then_set_cancellation_waiter_callback(loop, waiter):
    """
    Sets the cancellation waiter callback's result after 1 poll cycle.
    
    Parameters
    ----------
    loop : ``EventThread``
        Event loop to skip the cycle of.
    waiter : ``Future``
        Waiter to set.
    """
    loop.call_after(0.0, _set_cancellation_waiter_callback, waiter)


def _set_cancellation_waiter_callback(waiter):
    """
    Sets the cancellation waiter's callback's result.
    
    Parameters
    ----------
    waiter : ``Future``
        Waiter to set.
    """
    if waiter.set_result_if_pending(True):
        waiter._loop.wake_up()


async def _propagate_cancellation_async(future, loop):
    """
    Propagates cancellation and waits for result.
    
    This function is a coroutine.
    
    Parameters
    ----------
    future : ``Future``
        The future to cancel.
    loop : ``EventThread``
        Current event loop.
    
    Returns
    -------
    cancellation_exception : `None`, ``BaseException``
    """
    future_loop = future._loop
    if future.cancel():
        future_loop.wake_up()
    
    if future.is_cancelling():
        waiter = Future(loop)
        future_loop.call_soon(_skip_loop_then_set_cancellation_waiter_callback, future_loop, waiter)
        timeout_handle = loop.call_after(CANCELLATION_TIMEOUT, Future.set_result_if_pending, waiter, False)
        future_loop.wake_up()
        
        try:
            result_set = await waiter
        except:
            # Interrupted -> cancel timeout
            timeout_handle.cancel()
            raise
        
        # Not timeout -> cancel timeout
        if result_set:
            timeout_handle.cancel()
    
    # Use get exception in case an exception occurred while cancellation
    if future.is_done():
        return future.get_exception()


@export
class FutureWrapperAsync(FutureWrapperBase):
    """
    Async wrapper for ``Future``-s enabling them to be awaited from an another event loop.
    
    Attributes
    ----------
    _future : ``Future``
        The waited future. If the future's state == modified by the sync wrapper, then ``._future`` is set as `None`,
        to not retrieve the result again.
    
    _loop : ``EventThread``
        The loop to what the async future wrapper is bound to.
    """
    __slots__ = ('_loop',)
    
    def __new__(cls, future, loop):
        """
        Creates a new ``FutureWrapperAsync`` object bound to the given `loop` and `future`.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop from where the created wrapper futures can be awaited.
        future : ``Future``
            The future to wrap.
        """
        self = object.__new__(cls)
        self._future = future
        self._loop = loop
        return self
    
    
    async def wait(self, timeout = None, propagate_cancellation = False):
        """
        Waits till the waited future's result or exception is set.
        
        If the future is cancelled, raises ``CancelledError``.
        
        If the future has exception set with `.set_exception`, `.set_exception_if_pending` successfully, then raises
        the given exception.
        
        If the future has result set with `.set_result`, `.set_result_if_pending` successfully, then returns the
        given object.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional
            Timeout in seconds till the waited future's result should be set. Giving it as `None`, means no time limit.
        
        propagate_cancellation : `bool` = `False`, Optional
            Whether cancellation should be propagated towards the waited task.
        
        Returns
        -------
        result : `object`
            The future's result.
        
        Raises
        ------
        TimeoutError
            If `timeout` is over and the waited future is still pending.
        CancelledError
            The future is cancelled.
        TypeError
            The future has non `BaseException` set as exception.
        BaseException
            The future's set exception.
        """
        future = self._future
        if future.is_done():
            return future.get_result()
        
        waiter = Future(self._loop)
        
        callback = MethodType(_set_future_waiter_callback, waiter)
        future.add_done_callback(callback)
        
        if (timeout is None):
            timeout_handle = None
        else:
            timeout_handle = self._loop.call_after(timeout, Future.set_result_if_pending, waiter, False)
        
        future._loop.wake_up()
        
        try:
            result_set = await waiter
        except BaseException as err:
            # Interrupted -> remove callback
            future.remove_done_callback(callback)
            # Interrupted -> cancel timeout
            if (timeout_handle is not None):
                timeout_handle.cancel()
            
            # Cancel source future | Do not use await in GeneratorExit !!
            if propagate_cancellation:
                if isinstance(err, GeneratorExit):
                    future.cancel()
                else:
                    err.__cause__ = await _propagate_cancellation_async(future, self._loop)
            
            raise
        
        if result_set:
            return future.get_result()
        
        # Timeout -> remove callback
        future.remove_done_callback(callback)
        # Cancel source future
        if propagate_cancellation:
            cancellation_exception = await _propagate_cancellation_async(future, self._loop)
        else:
            cancellation_exception = None
        
        raise TimeoutError from cancellation_exception
    
    
    
    async def wait_for_completion(self, timeout = None, propagate_cancellation = False):
        """
        Waits till the waited future's result or exception is set.
        
        This method is a coroutine.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional
            Timeout in seconds till the waited future's result should be set. Giving it as `None`, means no time limit.
        
        propagate_cancellation : `bool` = `False`, Optional
            Whether cancellation should be propagated towards the waited task.
        
        Returns
        -------
        completed : `bool`
            Whether the wrapper future is completed.
        """
        future = self._future
        if future.is_done():
            return True
        
        waiter = Future(self._loop)
        
        callback = MethodType(_set_future_waiter_callback, waiter)
        future.add_done_callback(callback)
        
        if (timeout is None):
            timeout_handle = None
        else:
            timeout_handle = self._loop.call_after(timeout, Future.set_result_if_pending, waiter, False)
        
        future._loop.wake_up()
        
        try:
            result_set = await waiter
        except BaseException as err:
            # Interrupted -> remove callback
            future.remove_done_callback(callback)
            # Interrupted -> cancel timeout
            if (timeout_handle is not None):
                timeout_handle.cancel()
            
            # Cancel source future | Do not use await in GeneratorExit !!
            if propagate_cancellation:
                if isinstance(err, GeneratorExit):
                    future.cancel()
                else:
                    err.__cause__ = await _propagate_cancellation_async(future, self._loop)
            
            raise
        
        if result_set:
            return True
        
        # Timeout -> remove callback
        future.remove_done_callback(callback)
        # Cancel source future
        if propagate_cancellation:
            await _propagate_cancellation_async(future, self._loop)
        
        return False
    
    
    def __iter__(self):
        """Awaits the wrapped future."""
        future = self._future
        if future.is_done():
            return future.get_result()
        
        waiter = Future(self._loop)
        
        callback = MethodType(_set_future_waiter_callback, waiter)
        future.add_done_callback(callback)
        
        future._loop.wake_up()
        
        try:
            yield from waiter
        except:
            # Interrupted -> remove callback
            future.remove_done_callback(callback)
            raise
        
        return future.get_result()
    
    
    __await__ = __iter__



__all__ = ('check_satisfaction', 'export', 'include', 'include_with_callback')

import sys, warnings
from functools import partial as partial_func
from importlib.util import module_from_spec
from types import ModuleType


CALLBACKS = {}
SATISFIED = {}


def include(obj_name):
    """
    Includes the object with the given name. The requirement is satisfied when the respective object is exported.
    
    Parameters
    ----------
    obj_name : `str`
        The object name to include when exported.
    
    Returns
    -------
    placeholder : `value` / `NotImplementedType`
        Placeholder to trick the editors.
    """
    frame = sys._getframe().f_back
    spec = frame.f_globals.get('__spec__', None)
    if spec is None:
        module = None
    else:
        module = sys.modules.get(spec.name, None)
        if module is None:
            module = module_from_spec(spec)
    
    try:
        value = SATISFIED[obj_name]
    except KeyError:
        pass
    else:
        return value
    
    if (module is None):
        warnings.warn(
            (
                f'Cannot include `{obj_name!r}` into {frame.f_globals.get("__file__", None)}.\n'
                f'The object is not yet resolved & The file is not a module.'
            ),
            ResourceWarning,
            stacklevel = 2,
        )
    
    else:
        _add_callback(obj_name, partial_func(_include_callback, module, obj_name))
    
    return NotImplemented


def include_with_callback(obj_name, callback = None):
    """
    Includes the object with the given name. If the respective object is exported the given `callback` is called.
    
    Parameters
    ----------
    obj_name : `str`
        The object name to pass a parameter when exported.
    
    callback : `None`, `callable` = `None`, Optional
        The callback to run when the requirement is satisfied.
    
    Returns
    -------
    none / decorator : `None` / `functools.partial`
        Allows using it as a decorator if only `obj_name` was passed.
    """
    if callback is None:
        return partial_func(include_with_callback, obj_name)
    
    try:
        value = SATISFIED[obj_name]
    except KeyError:
        _add_callback(obj_name, callback)
    else:
        callback(value)


def export(obj, obj_name = None):
    """
    Exports the given object.
    
    Parameters
    ----------
    obj : `Any`
        The object to export.
    obj_name : `str` = `None`, Optional
        The name of the object. If not given, is detected from `obj` itself.
    
    Returns
    -------
    obj : `Any`
        The exported object.
    """
    if obj_name is None:
        try:
            obj_name = obj.__name__
        except AttributeError:
            obj_name = obj.__class__.__name__
    
    if isinstance(obj, ModuleType):
        obj_name = obj_name[obj_name.rfind('.') + 1:]
    
    SATISFIED[obj_name] = obj
    
    try:
        callbacks = CALLBACKS.pop(obj_name)
    except KeyError:
        pass
    else:
        for callback in callbacks:
            callback(obj)
    
    return obj


def _include_callback(module, obj_name, obj):
    """
    Callback ran by ``.include`` if the object cannot be resolved initially.
    
    Parameters
    ----------
    module : `ModuleType`
        Module to set `obj`.
    obj_name : `str`
        The object's name to set as.
    obj : `object`
        the object to set.
    """
    setattr(module, obj_name, obj)


def _add_callback(obj_name, callback):
    """
    Adds callback to run when an object is exported.
    
    Parameters
    ----------
    obj_name : `str`
        the object's name to add the resolve callback for.
    
    callback : `callable`
        The function to call when the export is resolved.
    """
    try:
        callbacks = CALLBACKS[obj_name]
    except KeyError:
        callbacks = []
        CALLBACKS[obj_name] = callbacks
    
    callbacks.append(callback)


def check_satisfaction():
    """
    Checks whether every ``include`` requirement is satisfied with ``export``.
    
    Raises
    ------
    AssertionError
    """
    assert (not CALLBACKS), f'Unsatisfied includes: {", ".join(CALLBACKS.keys())}'



__all__ = ('CauseGroup',)



def _iter_validate_causes(causes):
    """
    Validates the given causes.
    
    Parameters
    ----------
    causes : `tuple` of (`BaseException`, `type<BaseException>`)
        Exception causes.
    
    Raises
    ------
    TypeError
        If a cause is not `BaseException` instance.
    """
    for cause in causes:
        if isinstance(cause, BaseException):
            yield cause
            continue
        
        if isinstance(cause, type) and issubclass(cause, BaseException):
            yield cause()
            continue
        
        raise TypeError(
            f'Causes can be `{BaseException.__name__}`, got {cause.__class__.__name__}; {cause!r}.'
        )


class CauseGroup(BaseException):
    """
    Allows an exception to have multiple causes.
    
    Example
    -------
    ```py3
    raise Exception from CauseGroup(...)
    ```
    
    Attributes
    ----------
    causes : `tuple` of ``BaseException``
        Exception causes.
    """
    def __new__(cls, *causes):
        """
        Stops ``CauseGroup`` to be created if a cause is not `BaseException` instance.
        
        Parameters
        ----------
        *causes : `BaseException`, `type<BaseException>`
            Exception causes.
        
        Raises
        ------
        TypeError
            If a cause is not `BaseException` instance.
        
        Returns
        -------
        case : `None`, `BaseException`, `CauseGroup`
            Only returns a cause group if there are at least 2 exceptions to group.
        """
        cause_count = len(causes)
        
        if cause_count == 0:
            return None
        
        if cause_count == 1:
            return causes[0]
        
        causes = tuple(_iter_validate_causes(causes))
        
        self = BaseException.__new__(cls)
        self.causes = causes
        return self
    
    
    # What the hell is wrong with `str()` ?
    __str__ = BaseException.__repr__
    
    
    def __iter__(self):
        """
        Iterates over the causes of the cause group.
        
        This method is an iterable generator.
        
        Yields
        ------
        cause : `BaseException`
        """
        yield from self.causes
    
    
    def __len__(self):
        """
        Returns how much exception the cause group contains.
        
        Returns
        -------
        length : `int`
        """
        return len(self.causes)



__all__ = ()

BUILTINS = (
    'ArithmeticError', 'AssertionError', 'AttributeError', 'BaseException', 'BlockingIOError', 'BrokenPipeError',
    'BufferError', 'BytesWarning', 'ChildProcessError', 'ConnectionAbortedError', 'ConnectionError',
    'ConnectionRefusedError', 'ConnectionResetError', 'DeprecationWarning', 'EOFError', 'Ellipsis',
    'EnvironmentError', 'Exception', 'False', 'FileExistsError', 'FileNotFoundError', 'FloatingPointError',
    'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError', 'ImportWarning', 'IndentationError', 'IndexError',
    'InterruptedError', 'IsADirectoryError', 'KeyError', 'KeyboardInterrupt', 'LookupError', 'MemoryError',
    'ModuleNotFoundError', 'NameError', 'None', 'NotADirectoryError', 'NotImplemented', 'NotImplementedError',
    'OSError', 'OverflowError', 'PendingDeprecationWarning', 'PermissionError', 'ProcessLookupError',
    'RecursionError', 'ReferenceError', 'ResourceWarning', 'RuntimeError', 'RuntimeWarning', 'StopAsyncIteration',
    'StopIteration', 'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit', 'TabError', 'TimeoutError',
    'True', 'TypeError', 'UnboundLocalError', 'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError',
    'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning', 'ValueError', 'Warning', 'ZeroDivisionError',
    '__import__', 'abs', 'all', 'any', 'ascii', 'bin', 'bool', 'breakpoint', 'bytearray', 'bytes', 'callable',
    'chr', 'classmethod', 'compile', 'complex', 'copyright', 'credits', 'delattr', 'dict', 'dir', 'divmod',
    'enumerate', 'eval', 'exec', 'exit', 'filter', 'float', 'format', 'frozenset', 'getattr', 'globals', 'hasattr',
    'hash', 'help', 'hex', 'id', 'input', 'int', 'isinstance', 'issubclass', 'iter', 'len', 'license', 'list',
    'locals', 'map', 'max', 'memoryview', 'min', 'next', 'object', 'oct', 'open', 'ord', 'pow', 'print',
    'property', 'quit', 'range', 'repr', 'reversed', 'round', 'set', 'setattr', 'slice', 'sorted', 'staticmethod',
    'str', 'sum', 'super', 'tuple', 'type', 'vars', 'zip'
)


class AutoCompleter:
    """
    Auto completer used in editors.
    
    Attributes
    ----------
    variables : `dict<str, object>`
        Variables to auto-complete from.
    """
    __slots__ = ('_variables',)
    
    def __new__(cls, variables):
        """
        Creates a new editor auto completer.
        
        Parameters
        ----------
        variables : `dict<str, object>`
            Variables to auto-complete from.
        """
        self = object.__new__(cls)
        self._variables = variables
        return self
    
    
    def __repr__(self):
        """Returns the auto completer's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' variables: ')
        repr_parts.append(repr(len(self._variables)))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def get_common_prefix(self, string):
        """
        Gets the first matching option.
        
        Parameters
        ----------
        string : `str`
            The string to match.
        
        Returns
        -------
        match : `None`, `str`
        """
        matches = {*self._iter_matches(string)}
        matches_count = len(matches)
        
        if matches_count == 0:
            return None
        
        first_match = next(iter(matches))
        if matches_count == 1:
            return first_match
        
        shortest_match_length = min(len(match) for match in matches)
        
        index = len(string)
        
        while index < shortest_match_length:
            character = first_match[index]
            if not all(match[index] == character for match in matches):
                break
                
            index += 1
            continue
        
        if index == len(string):
            return None
        
        return first_match[:index]
    
    
    def _iter_matches(self, string):
        """
        Iterates over the variable names matching the given string.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        string : `str`
            The string to match.
        
        Yields
        ------
        matched_name : `str`
        """
        for variable_name in self._iter_variable_names():
            if variable_name.startswith(string) and variable_name != string:
                yield variable_name
    
    
    def _iter_variable_names(self):
        """
        Iterates over the variable names registered to the auto completer. Might yield duplicates.
        
        This method is an iterable generator.
        
        Yields
        ------
        variable_name : `str`
        """
        yield from self._variables.keys()
        yield from BUILTINS



__all__ = ()

from datetime import datetime as DateTime
from warnings import warn

from ...utils import copy_docs

from .future import Future
from .future_repr import render_callbacks_into, render_future_into, render_result_into, render_state_into


SILENCE_DEPRECATED =  DateTime.utcnow() > DateTime(2023, 11, 12)

CANCELLATION_TIMEOUT = 0.0004


class FutureWrapperBase:
    """
    Base future wrapper for cross-thread setting and retrieving.
    
    Attributes
    ----------
    _future : ``Future``
        The waited future.
    """
    __slots__ = ('_future',)
    
    def __new__(cls, future):
        """
        Creates a new future wrapper wrapping the given `future`.
        
        Parameters
        ----------
        future : ``Future``
            The future on what we would want to wait from a sync thread.
        """
        self = object.__new__(cls)
        self._future = future
        return self


    def __repr__(self):
        """Returns the future sync wrapper's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        future = self._future
        state = future._state
        repr_parts, field_added = render_state_into(repr_parts, False, state)
        repr_parts, field_added = render_result_into(repr_parts, field_added, state, future._result)
        repr_parts, field_added = render_future_into(repr_parts, field_added, future)
        repr_parts, field_added = render_callbacks_into(repr_parts, field_added, future._callbacks)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(Future.cancel)
    def cancel(self):
        future = self._future
        cancelled = future.cancel()
        if cancelled:
            future._loop.wake_up()
        return cancelled
    
    
    @copy_docs(Future.cancel_with)
    def cancel_with(self, exception):
        future = self._future
        cancelled = future.cancel_with(exception)
        if cancelled:
            future._loop.wake_up()
        return cancelled
    
    
    @copy_docs(Future.silence)
    def silence(self):
        self._future.silence()
    
    
    @copy_docs(Future.is_cancelled)
    def is_cancelled(self):
        return self._future.is_cancelled()
    
    
    @copy_docs(Future.is_cancelling)
    def is_cancelling(self):
        return self._future.is_cancelling()
    
    
    @copy_docs(Future.is_done)
    def is_done(self):
        return self._future.is_done()
    
    
    @copy_docs(Future.is_pending)
    def is_pending(self):
        return self._future.is_pending()
    
    
    @copy_docs(Future.get_result)
    def get_result(self):
        return self._future.get_result()
    
    
    @copy_docs(Future.get_exception)
    def get_exception(self):
        return self._future.get_exception()
    
    
    @copy_docs(Future.get_cancellation_exception)
    def get_cancellation_exception(self):
        return self._future.get_cancellation_exception()
    
    
    @copy_docs(Future.set_result)
    def set_result(self, result):
        future = self._future
        future.set_result(result)
        future._loop.wake_up()
    
    
    @copy_docs(Future.set_result_if_pending)
    def set_result_if_pending(self, result):
        future = self._future
        result_set = future.set_result_if_pending(result)
        if result_set:
            future._loop.wake_up()
        
        return result_set
    
    
    @copy_docs(Future.set_exception)
    def set_exception(self, exception):
        future = self._future
        future.set_exception(exception)
        future._loop.wake_up()
    
    
    @copy_docs(Future.set_exception_if_pending)
    def set_exception_if_pending(self, exception):
        future = self._future
        exception_set = future.set_exception_if_pending(exception)
        if exception_set:
            future._loop.wake_up()
        
        return exception_set
    
    
    def wait(self, timeout = None, propagate_cancellation = False):
        """
        Waits till the waited future's result or exception is set.
        
        If the future is cancelled, raises ``CancelledError``.
        
        If the future has exception set with `.set_exception`, `.set_exception_if_pending` successfully, then raises
        the given exception.
        
        If the future has result set with `.set_result`, `.set_result_if_pending` successfully, then returns the
        given object.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional
            Timeout in seconds till the waited future's result should be set. Giving it as `None`, means no time limit.
        
        propagate_cancellation : `bool` = `False`, Optional
            Whether cancellation should be propagated towards the waited task.
        
        Raises
        ------
        TimeoutError
            If `timeout` is over and the waited future is still pending.
        CancelledError
            The future is cancelled.
        TypeError
            The future has non `BaseException` set as exception.
        BaseException
            The future's set exception.
        """
        raise NotImplementedError
    
    
    def wait_for_completion(self, timeout = None, propagate_cancellation = False):
        """
        Waits till the waited future's result or exception is set.
        
        Parameters
        ----------
        timeout : `None`, `float` = `None`, Optional
            Timeout in seconds till the waited future's result should be set. Giving it as `None`, means no time limit.
        
        propagate_cancellation : `bool` = `False`, Optional
            Whether cancellation should be propagated towards the waited task.
        
        Returns
        -------
        completed : `bool`
            Whether the wrapper future is completed.
        """
        raise NotImplementedError
    
    # Deprecations
    
    if __debug__:
        def __silence__(self):
            """
            Silences the future's `__del__`, so it will not notify if it would.
            
            Deprecated and will be removed in 2024 February.
            
            Notes
            -----
            This method is present only if `__debug__` is set as `True`.
            """
            if SILENCE_DEPRECATED:
                warn(
                    (
                        f'`{self.__class__.__name__}.__silence__` is deprecated and will be removed in 2024 february.'
                        f'Please use `.silence()` instead.'
                    ),
                    FutureWarning,
                    stacklevel = 2,
                )
            
            self.silence()
    

    def cancelled(self):
        warn(
            f'`{self.__class__.__name__}.cancelled` is deprecated.',
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_cancelled()
    
    
    def done(self):
        warn(
            f'`{self.__class__.__name__}.done` is deprecated.',
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_done()
    
    
    def pending(self):
        warn(
            f'`{self.__class__.__name__}.pending` is deprecated.',
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_pending()
    
    
    @property
    def result(self):
        warn(
            (
                f'`{self.__class__.__name__}.result` is deprecated and will be removed in 2023 November. '
                f'Use `.get_result()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.get_result
    
    
    @property
    def exception(self):
        warn(
            (
                f'`{self.__class__.__name__}.result` is deprecated and will be removed in 2023 November. '
                f'Use `.get_result()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.get_exception



__all__ = ('CoroutineTypeProxy',)

from types import CoroutineType

from ..async_utils import is_coroutine

from .field_factories import proxy_method_factory, proxy_property_factory
from .proxy_base import ProxyBase


class CoroutineTypeProxy(ProxyBase):
    """
    Coroutine function proxy.
    
    Attributes
    ----------
    _overwrites : `None`, `dict` of (`str`, `object`) items
        Field overwrites.
    _proxied : `CoroutineType`
        The proxied coroutine function.
    """
    __slots__ = ()
    
    def __new__(cls, coroutine):
        """
        Creates a new coroutine proxy.
        
        Parameters
        ----------
        to_proxy : `instance<cls>`, `CoroutineType`
            The coroutine function ot proxy.
        """
        if isinstance(coroutine, cls):
            return coroutine._copy_self_deep()
        
        if is_coroutine(coroutine):
            self = object.__new__(cls)
            self._proxied = coroutine
            self._overwrites = None
            return self
        
        raise TypeError(
            f'`coroutine` can be either a coroutine or `{cls.__name__}`, '
            f'got {coroutine.__class__.__name__}.'
        )
    
    
    __qualname__ = CoroutineType.__qualname__
    
    __class__ = proxy_property_factory('__class__')
    __doc__ = proxy_property_factory('__doc__')
    __name__ = proxy_property_factory('__name__')
    
    __await__ = proxy_method_factory('__await__')
    __reduce__ = proxy_method_factory('__reduce__')
    __reduce_ex__ = proxy_method_factory('__reduce_ex__')
    __sizeof__ = proxy_method_factory('__sizeof__')
    cr_await = proxy_property_factory('cr_await')
    cr_code = proxy_property_factory('cr_code')
    cr_frame = proxy_property_factory('cr_frame')
    cr_origin = proxy_property_factory('cr_origin')
    cr_running = proxy_property_factory('cr_running')
    
    send = proxy_method_factory('send')
    throw = proxy_method_factory('throw')
    close = proxy_method_factory('close')
